"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.metrics import deserialize as deserialize
from synalinks.src.metrics import get as get
from synalinks.src.metrics import serialize as serialize
from synalinks.src.metrics.f_score_metrics import BinaryF1Score as BinaryF1Score
from synalinks.src.metrics.f_score_metrics import BinaryFBetaScore as BinaryFBetaScore
from synalinks.src.metrics.f_score_metrics import F1Score as F1Score
from synalinks.src.metrics.f_score_metrics import FBetaScore as FBetaScore
from synalinks.src.metrics.f_score_metrics import ListF1Score as ListF1Score
from synalinks.src.metrics.f_score_metrics import ListFBetaScore as ListFBetaScore
from synalinks.src.metrics.metric import Metric as Metric
from synalinks.src.metrics.reduction_metrics import Mean as Mean
from synalinks.src.metrics.reduction_metrics import MeanMetricWrapper as MeanMetricWrapper
from synalinks.src.metrics.reduction_metrics import Sum as Sum
from synalinks.src.metrics.regression_metrics import CosineSimilarity as CosineSimilarity



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import warnings

from synalinks.src import ops
from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import Embedding as EmbeddingVector
from synalinks.src.backend import JsonDataModel
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib
from synalinks.src.utils.async_utils import run_maybe_nested


@synalinks_export(
    [
        "synalinks.modules.EmbedKnowledge",
        "synalinks.EmbedKnowledge",
    ]
)
class EmbedKnowledge(Module):
    """Extracts a field of interest and generate the corresponding embedding vector.

    This module is designed to work with any data model structure. It supports to mask the
    entity fields in order to keep **only one** field to embed per data model.

    **Note**: Each data model should have the *same field* to compute the embedding
        from like a `name` or `description` field using `in_mask`.
        **Or** every data model should have *only one field left* after masking using
        `out_mask` argument.

    ```python
    import synalinks
    import asyncio
    from typing import Literal

    class Document(synalinks.DataModel):
        title: str = synalinks.Field(
            description="The document title",
        )
        text: str = synalinks.Field(
            description="The document content",
        )

    async def main():
        inputs = synalinks.Input(data_model=Document)
        outputs = await synalinks.EmbedKnowledge(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="embbed_document",
            description="Embbed the given documents"
        )

        doc = Document(
            title="my title",
            text="my document",
        )

        result = await program(doc)

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    If you want to process batch asynchronously
    use `program.predict()` instead, see the [FAQ](https://synalinks.github.io/synalinks/FAQ/#whats-the-difference-between-program-methods-predict-and-__call__)
    to understand the difference between `program()` and `program.predict()`

    Here is an example:

    ```python
    import synalinks
    import asyncio
    import numpy as np
    from typing import Literal

    class Document(synalinks.Entity):
        label: Literal["Document"]
        text: str = synalinks.Field(
            description="The document content",
        )

    async def main():
        inputs = synalinks.Input(data_model=Document)
        outputs = await synalinks.EmbedKnowledge(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="embbed_document",
            description="Embbed the given documents"
        )

        doc1 = Document(label="Document", text="my document 1")
        doc2 = Document(label="Document", text="my document 2")
        doc3 = Document(label="Document", text="my document 3")

        docs = np.array([doc1, doc2, doc3], dtype="object")

        embedded_docs = await program.predict(docs)

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    Args:
        embedding_model (EmbeddingModel): The embedding model to use.
        in_mask (list): A mask applied to keep specific entity fields.
        out_mask (list): A mask applied to remove specific entity fields.
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        embedding_model=None,
        in_mask=None,
        out_mask=None,
        name=None,
        description=None,
        trainable=False,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        self.embedding_model = embedding_model
        self.in_mask = in_mask
        self.out_mask = out_mask

    async def _embed(self, data_model):
        embeddings = data_model.get("embeddings")
        if embeddings:
            warnings.warn(
                "Embeddings already generated for this data model. "
                "Returning original data model."
            )
            return JsonDataModel(
                json=data_model.get_json(),
                schema=data_model.get_schema(),
                name="embedded_" + data_model.name,
            )
        masked_data_model = data_model
        if self.out_mask:
            masked_data_model = await ops.out_mask(
                data_model,
                mask=self.out_mask,
                recursive=False,
                name="out_mask_" + data_model.name,
            )
        elif self.in_mask:
            masked_data_model = await ops.in_mask(
                data_model,
                mask=self.in_mask,
                recursive=False,
                name="in_mask_" + data_model.name,
            )
        embeddings = await ops.embedding(
            masked_data_model,
            embedding_model=self.embedding_model,
            name=data_model.name + "_embedding",
        )
        if not embeddings or not embeddings.get("embeddings"):
            warnings.warn(
                f"No embeddings generated for data model {data_model.name}. "
                "Please check that your schema is correct."
            )
            return None
        embedding_list = embeddings.get("embeddings")
        if len(embedding_list) != 1:
            warnings.warn(
                "Data models can only have one embedding vector per data model, "
                "adjust `EmbedKnowledge` module's `in_mask` or `out_mask` "
                "to keep only one field. Skipping embedding."
            )
            return None
        vector = embedding_list[0]
        return await ops.concat(
            data_model,
            EmbeddingVector(embedding=vector),
            name="embedded_" + data_model.name,
        )

    async def call(self, inputs):
        if not inputs:
            return None
        return tree.map_structure(
            lambda x: run_maybe_nested(self._embed(x)),
            inputs,
        )

    async def compute_output_spec(self, inputs):
        return tree.map_structure(
            lambda x: x.clone(name="embedded_" + x.name),
            inputs,
        )

    def get_config(self):
        config = {
            "in_mask": self.in_mask,
            "out_mask": self.out_mask,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        embedding_model_config = {
            "embedding_model": serialization_lib.serialize_synalinks_object(
                self.embedding_model
            )
        }
        return {**embedding_model_config, **config}

    @classmethod
    def from_config(cls, config):
        embedding_model = serialization_lib.deserialize_synalinks_object(
            config.pop("embedding_model")
        )
        return cls(embedding_model=embedding_model, **config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.modules import Input
from synalinks.src.modules.masking.in_mask import InMask
from synalinks.src.programs import Program


class Document(DataModel):
    title: str
    text: str


class InMaskTest(testing.TestCase):
    async def test_in_mask_single_data_model(self):
        inputs = Input(data_model=Document)

        outputs = await InMask(
            mask=["text"],
        )(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
            name="masking_program",
            description="A program to keep fields",
        )

        doc = Document(title="Test document", text="Hello world")

        result = await program(doc)

        self.assertTrue(len(result.keys()) == 1)

    async def test_in_mask_multiple_data_models(self):
        inputs = [Input(data_model=Document), Input(data_model=Document)]

        outputs = await InMask(
            mask=["text"],
        )(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
            name="masking_program",
            description="A program to keep fields",
        )

        doc1 = Document(title="Test document 1", text="Hello world")
        doc2 = Document(title="Test document 2", text="Hello world")

        results = await program([doc1, doc2])

        self.assertTrue(len(results[0].keys()) == 1)
        self.assertTrue(len(results[1].keys()) == 1)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.api.ops import embedding_models as embedding_models
from synalinks.api.ops import json as json
from synalinks.api.ops import knowledge_bases as knowledge_bases
from synalinks.src.backend import convert_to_json_data_model as convert_to_json_data_model
from synalinks.src.backend import (
    convert_to_symbolic_data_model as convert_to_symbolic_data_model,
)
from synalinks.src.ops.embedding_models import embedding as embedding
from synalinks.src.ops.json import concat as concat
from synalinks.src.ops.json import concat as concatenate
from synalinks.src.ops.json import factorize as factorize
from synalinks.src.ops.json import in_mask as in_mask
from synalinks.src.ops.json import logical_and as logical_and
from synalinks.src.ops.json import logical_not as logical_not
from synalinks.src.ops.json import logical_or as logical_or
from synalinks.src.ops.json import logical_xor as logical_xor
from synalinks.src.ops.json import out_mask as out_mask
from synalinks.src.ops.json import prefix as prefix
from synalinks.src.ops.json import suffix as suffix
from synalinks.src.ops.knowledge_bases import similarity_search as similarity_search
from synalinks.src.ops.knowledge_bases import update_knowledge as update_knowledge
from synalinks.src.ops.language_models import predict as predict



# Modified from: keras/src/backend/common/keras_tensor.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import orjson

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend.common.json_schema_utils import is_schema_equal
from synalinks.src.backend.common.json_schema_utils import standardize_schema
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable
from synalinks.src.utils.async_utils import run_maybe_nested
from synalinks.src.utils.naming import auto_name


@synalinks_export("synalinks.SymbolicDataModel")
class SymbolicDataModel(SynalinksSaveable):
    """A symbolic backend-independent data model.

    A `SymbolicDataModel` is a container for a JSON schema and can be used to represent
        data structures in a backend-agnostic way. It can record history and is used in
        symbolic operations (in the Functional API and to compute output specs).

    A "symbolic data model" can be understood as a placeholder for data specification,
        it does not contain any actual data, only a schema. It can be used for building
        Functional models, but it cannot be used in actual computations.

    Args:
        data_model (DataModel): Optional. The data_model used to extract the schema.
        schema (dict): Optional. The JSON schema to be used. If the schema is not
            provided, the data_model argument should be used to infer it.
        record_history (bool): Optional. Boolean indicating if the history
            should be recorded. Defaults to `True`.
        name (str): Optional. A unique name for the data model. Automatically generated
            if not set.

    Examples:

    **Creating a `SymbolicDataModel` with a backend data model metaclass:**

    ```python
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    data_model = SymbolicDataModel(data_model=Query)
    ```

    **Creating a `SymbolicDataModel` with a backend data model metaclass's schema:**

    ```python
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    data_model = SymbolicDataModel(schema=Query.get_schema())
    ```

    **Creating a `SymbolicDataModel` with `to_symbolic_data_model()`:**

    using a backend data model metaclass

    ```python
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    data_model = Query.to_symbolic_data_model()
    ```
    """

    def __init__(
        self,
        data_model=None,
        schema=None,
        record_history=True,
        name=None,
    ):
        self.name = name or auto_name(self.__class__.__name__)
        self._record_history = record_history
        self._schema = None
        if not schema and not data_model:
            raise ValueError(
                "You should specify at least one argument between "
                "`data_model` or `schema`"
            )
        if schema and data_model:
            if not is_schema_equal(schema, data_model.get_schema()):
                raise ValueError(
                    "Attempting to create a SymbolicDataModel "
                    "with both `schema` and `data_model` argument "
                    "but their schemas are incompatible "
                    f"received schema={schema} and "
                    f"data_model.get_schema()={data_model.get_schema()}."
                )
            self._schema = standardize_schema(schema)
        else:
            if schema:
                self._schema = standardize_schema(schema)
            if data_model:
                self._schema = standardize_schema(data_model.get_schema())

    @property
    def record_history(self):
        """Whether the history is being recorded."""
        return self._record_history

    @record_history.setter
    def record_history(self, value):
        self._record_history = value

    def get_schema(self):
        """Gets the JSON schema of the data model.

        Returns:
            (dict): The JSON schema.
        """
        return self._schema

    def get_json(self):
        """Gets the current value of the JSON object (impossible in `SymbolicDataModel`).

        Implemented to help the user to identifying issues.

        Raises:
            ValueError: The help message.
        """
        raise ValueError(
            "Attempting to retrieve the JSON value from a symbolic data model "
            "this operation is not possible, make sure that your `call()` "
            "is correctly implemented, if so then you likely need to implement "
            " `compute_output_spec()` in your subclassed module."
        )

    def prettify_schema(self):
        """Get a pretty version of the JSON schema for display.

        Returns:
            (dict): The indented JSON schema.
        """
        return orjson.dumps(self._schema, option=orjson.OPT_INDENT_2).decode()

    def __repr__(self):
        return f"<SymbolicDataModel schema={self._schema}, name={self.name}>"

    def __add__(self, other):
        """Concatenates this data model with another.

        Args:
            other (SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (SymbolicDataModel): The concatenated data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Concat().symbolic_call(self, other))

    def __radd__(self, other):
        """Concatenates (reverse) another data model with this one.

        Args:
            other (SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (SymbolicDataModel): The concatenated data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Concat().symbolic_call(other, self))

    def __and__(self, other):
        """Perform a `logical_and` with another data model.

        If one of them is None, output None. If both are provided,
        then concatenates this data model with the other.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): The concatenated data model or None
                based on the `logical_and` table.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.And().symbolic_call(self, other))

    def __rand__(self, other):
        """Perform a `logical_and` (reverse) with another data model.

        If one of them is None, output None. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): The concatenated data model or None
                based on the `logical_and` table.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.And().symbolic_call(other, self))

    def __or__(self, other):
        """Perform a `logical_or` with another data model

        If one of them is None, output the other one. If both are provided,
        then concatenates this data model with the other.

        Args:
            other (SymbolicDataModel): The other data model to concatenate with.

        Returns:
            (SymbolicDataModel | None): The concatenation of data model if both are
                provided, or the non-None data model or None if none are provided.
                (See `logical_or` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Or().symbolic_call(self, other))

    def __ror__(self, other):
        """Perform a `logical_or` (reverse) with another data model

        If one of them is None, output the other one. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): The concatenation of data model if both are
                provided, or the non-None data model or None if none are provided.
                (See `logical_or` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Or().symbolic_call(other, self))

    def __xor__(self, other):
        """Perform a `logical_xor` with another data model.

        If one of them is `None`, output the other one. If both are provided,
        then the output is `None`.

        Args:
            other (SymbolicDataModel): The other data model to concatenate with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Xor().symbolic_call(self, other))

    def __rxor__(self, other):
        """Perform a `logical_xor` (reverse) with another data model.

        If one of them is None, output the other one. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Xor().symbolic_call(other, self))

    def __contains__(self, other):
        """Check if the schema of `other` is contained in this one,
        or if a string key exists.

        Args:
            other (SymbolicDataModel | DataModel | str): The other data model to compare
                with, or a string key to check for in the schema properties.

        Returns:
            (bool): True if all properties of `other` are present in this one,
                or if the string key exists in the schema properties.
        """
        if isinstance(other, str):
            schema = self.get_schema()
            return other in schema.get("properties", {})
        from synalinks.src.backend.common.json_schema_utils import contains_schema

        return contains_schema(self.get_schema(), other.get_schema())

    def __invert__(self):
        """Perform an invertion/negation

        When an input is provided, invert it by outputing `None`

        Returns:
            (SymbolicDataModel | None): `None` if used with an instance/class,
                and a symbolic data model if used on a metaclass or symbolic model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Not().symbolic_call(self))

    def factorize(self):
        """Factorizes the data model.

        Returns:
            (SymbolicDataModel): The factorized data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Factorize().symbolic_call(self))

    def in_mask(self, mask=None, recursive=True):
        """Applies a mask to **keep only** specified keys of the data model.

        Args:
            mask (list): The mask to be applied (list of keys).
            recursive (bool): Optional. Whether to apply the mask recursively.
                Defaults to `True`.

        Returns:
            (SymbolicDataModel): The data model with the mask applied.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.InMask(mask=mask, recursive=True).symbolic_call(self))

    def out_mask(self, mask=None, recursive=True):
        """Applies an mask to **remove** specified keys of the data model.

        Args:
            mask (list): The mask to be applied (list of keys).
            recursive (bool): Optional. Whether to apply the mask recursively.
                Defaults to `True`.

        Returns:
            (SymbolicDataModel): The data model with the mask applied.
        """
        from synalinks.src import ops

        return run_maybe_nested(
            ops.OutMask(mask=mask, recursive=True).symbolic_call(self)
        )

    def prefix(self, prefix=None):
        """Add a prefix to **all** the data model fields (non-recursive).

        Args:
            prefix (str): the prefix to add

        Returns:
            (SymbolicDataModel): The data model with the prefix added.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Prefix(prefix=prefix).symbolic_call(self))

    def suffix(self, suffix=None):
        """Add a suffix to **all** the data model fields (non-recursive).

        Args:
            suffix (str): the suffix to add

        Returns:
            (SymbolicDataModel): The data model with the suffix added.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Suffix(suffix=suffix).symbolic_call(self))

    def get(self, key, default=None):
        """Get wrapper to make easier to access fields.

        Implemented to help the user to identifying issues.

        Args:
            key (str): The key to access.
            default (any): The default value if key not found.

        Raises:
            ValueError: The help message.
        """
        raise ValueError(
            f"Attempting to get '{key}' from a symbolic data model "
            "this operation is not possible, make sure that your `call()` "
            "is correctly implemented, if so then you likely need to implement "
            " `compute_output_spec()` in your subclassed module."
        )

    def __getitem__(self, key):
        """Get item wrapper to make it easier to access JSON fields.

        Implemented to help the user to identifying issues.

        Args:
            key (str): The key to access.
        """
        raise ValueError(
            f"Attempting to get '{key}' from a symbolic data model "
            "this operation is not possible, make sure that your `call()` "
            "is correctly implemented, if so then you likely need to implement "
            " `compute_output_spec()` in your subclassed module."
        )

    def keys(self):
        """Keys wrapper to make it easier to access JSON fields."""
        if "properties" in self.get_schema():
            return self.get_schema()["properties"].keys()
        else:
            return []

    def values(self):
        """Values wrapper to make it easier to access JSON fields.

        Implemented to help the user to identifying issues.
        """
        raise ValueError(
            "Attempting to get '.values()' from a symbolic data model "
            "this operation is not possible, make sure that your `call()` "
            "is correctly implemented, if so then you likely need to implement "
            " `compute_output_spec()` in your subclassed module."
        )

    def items(self):
        """Items wrapper to make it easier to access JSON fields.

        Implemented to help the user to identifying issues.
        """
        raise ValueError(
            "Attempting to get '.items()' from a symbolic data model "
            "this operation is not possible, make sure that your `call()` "
            "is correctly implemented, if so then you likely need to implement "
            " `compute_output_spec()` in your subclassed module."
        )

    def update(self, kv_dict):
        """Update wrapper to make easier to modify fields.

        Implemented to help the user to identifying issues.

        Args:
            kv_dict (dict): The key/value dict to update.

        Raises:
            ValueError: The help message.
        """
        raise ValueError(
            f"Attempting to update keys {list(kv_dict.keys())} from a symbolic "
            "data model this operation is not possible, make sure that your `call()` "
            "is correctly implemented, if so then you likely need to implement "
            " `compute_output_spec()` in your subclassed module."
        )

    def clone(self, name=None):
        """Clone a symbolic data model and give it a different name."""
        import copy

        clone = copy.deepcopy(self)
        if name:
            clone.name = name
        else:
            clone.name = "clone_" + self.name
        return clone

    def get_config(self):
        config = {
            "name": self.name,
            "schema": self.get_schema(),
            "record_history": self.record_history,
        }
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def any_symbolic_data_models(args=None, kwargs=None):
    """Checks if any of the arguments are symbolic data models.

    Args:
        args (tuple): Optional. The positional arguments to check.
        kwargs (dict): Optional. The keyword arguments to check.

    Returns:
        (bool): True if any of the arguments are symbolic data models, False otherwise.
    """
    args = args or ()
    kwargs = kwargs or {}
    for x in tree.flatten((args, kwargs)):
        if is_symbolic_data_model(x):
            return True
    return False


@synalinks_export(
    [
        "synalinks.utils.is_symbolic_data_model",
        "synalinks.backend.is_symbolic_data_model",
    ]
)
def is_symbolic_data_model(x):
    """Returns whether `x` is a synalinks data model.

    A "synalinks data model" is a *symbolic data model*, such as a data model
    that was created via `Input()`. A "symbolic data model"
    can be understood as a placeholder for data specification -- it does not
    contain any actual data, only a schema.
    It can be used for building Functional models, but it
    cannot be used in actual computations.

    Args:
        x (any): The object to check.

    Returns:
        (bool): True if `x` is a symbolic data model, False otherwise.
    """
    return isinstance(x, SymbolicDataModel)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.ops.knowledge_bases import similarity_search as similarity_search
from synalinks.src.ops.knowledge_bases import update_knowledge as update_knowledge



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

from synalinks.src import optimizers
from synalinks.src import rewards
from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules import Generator
from synalinks.src.modules import Input
from synalinks.src.modules import default_instructions
from synalinks.src.programs import Program
from synalinks.src.utils.nlp_utils import remove_numerical_suffix


class ProgramTest(testing.TestCase):
    async def test_get_state(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama/mistral")

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        data_model_fields = list(AnswerWithRationale.get_schema()["properties"].keys())
        instructions = default_instructions(data_model_fields)

        state_tree = program.get_state_tree()

        expected_tree = {
            "trainable_variables": {
                "generator": {
                    "state_generator": {
                        "instructions": instructions,
                        "examples": [],
                        "current_predictions": [],
                        "predictions": [],
                        "history": [],
                        "seed_candidates": [],
                        "candidates": [],
                        "best_candidates": [],
                        "nb_visit": 0,
                        "cumulative_reward": 0.0,
                    }
                }
            },
            "non_trainable_variables": {},
            "metrics_variables": {},
        }
        self.assertEqual(state_tree, expected_tree)

    async def test_recover_state(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama/mistral")

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        state_tree = program.get_state_tree()
        state_tree["trainable_variables"]["generator"]["state_generator"][
            "prompt_template"
        ] = "Dummy prompt template"
        program.set_state_tree(state_tree)
        self.assertEqual(state_tree, program.get_state_tree())

    async def test_get_state_after_compile(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(model="ollama/mistral")

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        program.compile(
            reward=rewards.ExactMatch(),
            optimizer=optimizers.random_few_shot.RandomFewShot(),
        )

        state_tree = program.get_state_tree()

        data_model_fields = list(AnswerWithRationale.get_schema()["properties"].keys())
        instructions = default_instructions(data_model_fields)

        expected_tree = {
            "trainable_variables": {
                "generator": {
                    "state_generator": {
                        "instructions": instructions,
                        "examples": [],
                        "current_predictions": [],
                        "predictions": [],
                        "history": [],
                        "seed_candidates": [],
                        "candidates": [],
                        "best_candidates": [],
                        "nb_visit": 0,
                        "cumulative_reward": 0.0,
                    }
                }
            },
            "non_trainable_variables": {},
            "optimizer_non_trainable_variables": {
                "random_few_shot": {
                    "iterations_random_few_shot": {
                        "iterations": 0,
                        "epochs": 0,
                    },
                },
            },
            "optimizer_trainable_variables": {},
            "metrics_variables": {
                "reward": {"total_with_count": {"total": 0.0, "count": 0}}
            },
        }
        self.maxDiff = None
        self.assertEqual(state_tree, expected_tree)

    async def test_recover_state_after_compile(self):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        program.compile(
            reward=rewards.ExactMatch(),
            optimizer=optimizers.random_few_shot.RandomFewShot(),
        )

        state_tree = program.get_state_tree()
        state_tree["trainable_variables"]["generator"]["state_generator"][
            "prompt_template"
        ] = "Dummy prompt template"
        program.set_state_tree(state_tree)
        self.assertEqual(state_tree, program.get_state_tree())

    @patch("litellm.acompletion")
    async def test_get_state_after_training(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        program.compile(
            reward=rewards.ExactMatch(in_mask=["answer"]),
            optimizer=optimizers.random_few_shot.RandomFewShot(),
        )

        (x_train, y_train), (x_test, y_test) = testing.test_utils.load_test_data()

        mock_completion.side_effect = testing.test_utils.mock_completion_data()

        _ = await program.fit(
            x=x_train,
            y=y_train,
            batch_size=32,
        )

        _ = program.get_state_tree()

    @patch("litellm.acompletion")
    async def test_recover_state_after_training(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        program.compile(
            reward=rewards.ExactMatch(in_mask=["answer"]),
            optimizer=optimizers.random_few_shot.RandomFewShot(),
        )

        (x_train, y_train), (x_test, y_test) = testing.test_utils.load_test_data()

        mock_completion.side_effect = testing.test_utils.mock_completion_data()

        _ = await program.fit(
            x=x_train,
            y=y_train,
            batch_size=32,
        )

        state_tree = program.get_state_tree()
        program.set_state_tree(state_tree)
        new_state_tree = program.get_state_tree()
        self.assertEqual(state_tree, new_state_tree)

    @patch("litellm.acompletion")
    async def test_saving_after_training(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await Generator(
            data_model=AnswerWithRationale,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

        program.compile(
            reward=rewards.ExactMatch(in_mask=["answer"]),
            optimizer=optimizers.random_few_shot.RandomFewShot(),
        )

        (x_train, y_train), (x_test, y_test) = testing.test_utils.load_test_data()

        mock_completion.side_effect = testing.test_utils.mock_completion_data()

        _ = await program.fit(
            x=x_train,
            y=y_train,
            batch_size=32,
        )

        filepath = "/tmp/program.json"
        program.save("/tmp/program.json")
        cloned_program = Program.load(filepath)

        for var1 in program.variables:
            for var2 in cloned_program.variables:
                if remove_numerical_suffix(var1.path) == remove_numerical_suffix(
                    var2.path
                ):
                    self.assertEqual(var1.get_json(), var2.get_json())



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.api.datasets import arcagi as arcagi
from synalinks.api.datasets import gsm8k as gsm8k
from synalinks.api.datasets import hotpotqa as hotpotqa



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import inspect
import logging
import typing

import docstring_parser
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib

JsonSchema = typing.Union[
    typing.Dict[str, typing.Any],
    typing.List[typing.Any],
    str,
    int,
    float,
    bool,
    None,
]


def json_schema_type(py_type: typing.Any) -> JsonSchema:
    """Convert a Python type to a JSON schema type."""
    mapping = {
        int: "integer",
        float: "number",
        bool: "boolean",
        str: "string",
        type(None): "null",
    }

    # Check if type is a basic type
    if py_type in mapping:
        return mapping[py_type]

    # For unparameterized list and dict types
    if py_type is list or py_type is typing.List:
        return {"type": "array", "items": {}}
    if py_type is dict or py_type is typing.Dict:
        return {"type": "object", "additionalProperties": {}}

    origin = typing.get_origin(py_type)
    args = typing.get_args(py_type)

    if origin is typing.Union:
        # Handle Optional[type] which is Union[type, None]
        if len(args) == 2 and type(None) in args:
            return json_schema_type(args[0])
        else:
            return [json_schema_type(arg) for arg in args]

    if origin is list or origin is typing.List:
        schema_type = json_schema_type(args[0])
        if isinstance(schema_type, dict):
            return {"type": "array", "items": schema_type}
        return {
            "type": "array",
            "items": {"type": json_schema_type(args[0])},
        }

    if origin is dict or origin is typing.Dict:
        schema_type = json_schema_type(args[1])
        if isinstance(schema_type, dict):
            return {"type": "object", "additionalProperties": schema_type}
        return {"type": "object", "additionalProperties": {"type": schema_type}}

    raise ValueError(f"Cannot convert {py_type} to a JSON schema type")


def get_param_schema(
    param_name: str,
    param: inspect.Parameter,
    type_hints: typing.Dict[str, typing.Any],
    doc_parsed: docstring_parser.Docstring,
) -> JsonSchema:
    """Create a schema for a single parameter."""
    if param_name not in type_hints:
        raise ValueError(f"Missing type hint for parameter '{param_name}'")
    param_type = type_hints[param_name]
    param_type_str = json_schema_type(param_type)
    descriptions = (p.description for p in doc_parsed.params if p.arg_name == param_name)
    param_doc = next(descriptions, None)
    if param_doc is None:
        raise ValueError(f"Missing description for parameter '{param_name}' in docstring")

    param_schema = {}
    param_schema["description"] = param_doc.replace("\n", " ")
    param_schema["title"] = param_name.title().replace("_", " ")
    if isinstance(param_type_str, dict):
        param_schema.update(**param_type_str)
    else:
        param_schema["type"] = param_type_str

    if param.default is not param.empty:
        param_schema["default"] = param.default

    return param_schema


@synalinks_export(["synalinks.modules.Tool", "synalinks.Tool"])
class Tool(Module):
    """A module that wraps an async function as a callable tool.

    The `Tool` module allows you to wrap any async function and use it as a
    module within Synalinks programs. It automatically extracts the function's
    schema from its type hints and docstring.

    Example:

    ```python
    import synalinks

    @synalinks.saving.register_synalinks_serializable()
    async def calculate(expression: str):
        \"\"\"Calculate the result of a mathematical expression.

        Args:
            expression (str): The mathematical expression to calculate.
        \"\"\"
        result = eval(expression)
        return {"result": result}

    tool = synalinks.Tool(calculate)
    result = await tool(expression="2 + 2")
    ```

    Important:
        **No Optional Parameters**: All function parameters must be required.
        Optional parameters with default values are not supported because
        OpenAI and other providers require all parameters to be required
        in their structured output JSON schemas.

        **Complete Docstring Required**: The wrapped function must have a
        complete docstring with an `Args:` section that documents every
        parameter. The Tool extracts parameter descriptions from the docstring
        to build the JSON schema sent to the language model. Missing
        descriptions will raise a ValueError.

        Example of a properly documented tool function:

        ```python
        async def search(query: str, limit: int):
            \"\"\"Search the database for matching records.

            Args:
                query (str): The search query string.
                limit (int): Maximum number of results to return.
            \"\"\"
            # Implementation here
            return {"results": [...]}
        ```

    Args:
        func (Callable): The async function to wrap as a tool.
        name (str): Optional. The name of the module. Defaults to the function name.
        description (str): Optional. The description of the module.
            Defaults to the function's docstring short description.
        trainable (bool): Whether the module's variables should be trainable.
            Defaults to False since tools typically don't have trainable state.
    """

    def __init__(
        self,
        func: typing.Callable,
        name=None,
        description=None,
        trainable=False,
    ):
        self._func = func
        if not inspect.iscoroutinefunction(self._func):
            raise TypeError(f"{self._func.__name__} is not an asynchronous function")

        doc = inspect.getdoc(func)
        if not doc:
            raise ValueError(f"The tool ({self._func.__name__}) must have a docstring")

        self._docstring = docstring_parser.parse(doc)
        self._signature = inspect.signature(func)
        self._type_hints = typing.get_type_hints(func)
        self._params_schema = {}
        self._required_params = []

        self._parse_arguments()

        # Use function name if no name provided
        if not name:
            name = self._func.__name__

        # Use docstring short description if no description provided
        if not description:
            description = self._docstring.short_description or ""

        if not description:
            logging.warning(
                f"The tool ({name}) has no description. "
                "This is unsafe behavior and may lead to issues."
            )

        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )

    def _parse_arguments(self):
        """Parse the function arguments to build the input parameter schema."""
        for param_name, param in self._signature.parameters.items():
            param_schema = get_param_schema(
                param_name,
                param,
                self._type_hints,
                self._docstring,
            )
            self._params_schema[param_name] = param_schema
            if param.default is param.empty:
                self._required_params.append(param_name)

    def _build_output_schema(self):
        """Build the output schema from the function's return type hint.

        Since tools must always return a dict, this method ensures
        the output schema is always of type "object".
        """
        return_type = self._type_hints.get("return", None)
        base_schema = {
            "type": "object",
            "title": f"{self.name}_output",
        }

        if return_type is None:
            # No return type hint, use generic object schema
            base_schema["additionalProperties"] = True
            return base_schema

        origin = typing.get_origin(return_type)
        args = typing.get_args(return_type)

        # Handle Dict[K, V] - extract value type for additionalProperties
        if origin is dict or origin is typing.Dict:
            if len(args) >= 2:
                value_type = args[1]
                try:
                    value_schema = json_schema_type(value_type)
                    if isinstance(value_schema, dict):
                        base_schema["additionalProperties"] = value_schema
                    else:
                        base_schema["additionalProperties"] = {"type": value_schema}
                except ValueError:
                    base_schema["additionalProperties"] = True
            else:
                base_schema["additionalProperties"] = True
            return base_schema

        # Handle TypedDict - extract properties from annotations
        if hasattr(return_type, "__annotations__"):
            properties = {}
            required = []
            for field_name, field_type in return_type.__annotations__.items():
                try:
                    field_schema = json_schema_type(field_type)
                    if isinstance(field_schema, dict):
                        properties[field_name] = field_schema
                    else:
                        properties[field_name] = {"type": field_schema}
                    # Check if field is required (not Optional)
                    if typing.get_origin(field_type) is not typing.Union:
                        required.append(field_name)
                    elif type(None) not in typing.get_args(field_type):
                        required.append(field_name)
                except ValueError:
                    properties[field_name] = {}
            if properties:
                base_schema["properties"] = properties
                if required:
                    base_schema["required"] = required
                base_schema["additionalProperties"] = False
                return base_schema

        # Fallback to generic object schema
        base_schema["additionalProperties"] = True
        return base_schema

    def get_input_schema(self) -> dict:
        """Get the JSON schema for this tool's input parameters.

        Returns:
            dict: The JSON schema describing the tool's input parameters.
        """
        return {
            "additionalProperties": False,
            "description": self._docstring.short_description,
            "properties": self._params_schema,
            "required": self._required_params,
            "title": self.name.title().replace("_", " "),
            "type": "object",
        }

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_exception_type((Exception,)),
        reraise=True,
    )
    async def call(
        self, training: bool = False, **kwargs: typing.Any
    ) -> typing.Optional[JsonDataModel]:
        """Execute the wrapped function with the provided arguments.

        Args:
            training (bool): Whether in training mode. Not used by Tool but
                included for consistency with other modules.
            **kwargs (Any): The arguments to pass to the wrapped function.

        Returns:
            JsonDataModel: The result wrapped in a JsonDataModel with the output schema.
        """
        result = await self._func(**kwargs)
        if result is None:
            return None
        if isinstance(result, dict):
            return JsonDataModel(
                json=result,
                schema=self._build_output_schema(),
                name=f"{self.name}_output",
            )
        # Wrap non-dict results in a dict
        return JsonDataModel(
            json={"result": result},
            schema=self._build_output_schema(),
            name=f"{self.name}_output",
        )

    async def compute_output_spec(
        self, training: bool = False, **kwargs: typing.Any
    ) -> SymbolicDataModel:
        """Compute the output specification for the tool.

        Uses the function's schema to define the output structure.

        Args:
            training (bool): Whether in training mode.
            **kwargs (Any): The input arguments.

        Returns:
            SymbolicDataModel: A SymbolicDataModel with the tool's output schema.
        """
        return SymbolicDataModel(
            schema=self._build_output_schema(),
            name=self.name,
        )

    def get_tool_schema(self) -> dict:
        """Get the JSON schema for this tool's parameters.

        Returns:
            dict: The JSON schema describing the tool's input parameters.
        """
        schema = {
            "additionalProperties": False,
            "description": self._docstring.short_description,
            "properties": self._params_schema,
            "required": self._required_params,
            "title": self.name.title().replace("_", " "),
            "type": "object",
        }
        return schema

    def get_config(self):
        config = {
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        func_config = {"func": serialization_lib.serialize_synalinks_object(self._func)}
        return {**config, **func_config}

    @classmethod
    def from_config(cls, config):
        func = serialization_lib.deserialize_synalinks_object(config.pop("func"))
        return cls(func=func, **config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import platform
import unittest

import httpx
from mcp.server import FastMCP
from mcp.types import ToolAnnotations

from synalinks.src import testing
from synalinks.src.utils.mcp.client import MultiServerMCPClient
from synalinks.src.utils.mcp.test_common import run_streamable_server_multiprocessing


@unittest.skipUnless(
    platform.system() == "Linux",
    "server tests require Linux (multiprocessing/pickling issues on Windows)",
)
class MCPClientIntegrationTest(testing.TestCase):
    """integration tests for MultiServerMCPClient with actual MCP servers."""

    @classmethod
    def setUpClass(cls):
        """Set up multiple MCP servers for integration testing."""
        # server 1: time server with annotations
        cls.time_server = FastMCP(port=8181)

        @cls.time_server.tool(
            annotations=ToolAnnotations(
                title="Get Time", readOnlyHint=True, idempotentHint=False
            )
        )
        def get_time() -> str:
            """Get current time"""
            return "5:20:00 PM EST"

        # server 2: status server for custom HTTPX client testing
        cls.status_server = FastMCP(port=8182)

        @cls.status_server.tool()
        def get_status() -> str:
            """Get server status"""
            return "Server is running"

        @cls.status_server.tool()
        def get_uptime() -> str:
            """Get server uptime"""
            return "24h 30m 15s"

        # server 3: math server for multi-server testing
        cls.math_server = FastMCP(port=8183)

        @cls.math_server.tool()
        def add_numbers(a: int, b: int) -> int:
            """Add two numbers together"""
            return a + b

        @cls.math_server.tool()
        def multiply_numbers(a: int, b: int) -> int:
            """Multiply two numbers together"""
            return a * b

        cls.time_server_context = run_streamable_server_multiprocessing(cls.time_server)
        cls.status_server_context = run_streamable_server_multiprocessing(
            cls.status_server
        )
        cls.math_server_context = run_streamable_server_multiprocessing(cls.math_server)

        cls.time_server_context.__enter__()
        cls.status_server_context.__enter__()
        cls.math_server_context.__enter__()

        cls.time_connection = {
            "url": "http://localhost:8181/mcp/",
            "transport": "streamable_http",
        }

        cls.status_connection = {
            "url": "http://localhost:8182/mcp/",
            "transport": "streamable_http",
        }

        cls.math_connection = {
            "url": "http://localhost:8183/mcp/",
            "transport": "streamable_http",
        }

    @classmethod
    def tearDownClass(cls):
        """Tear down all MCP servers."""
        try:
            cls.time_server_context.__exit__(None, None, None)
        except Exception:
            pass

        try:
            cls.status_server_context.__exit__(None, None, None)
        except Exception:
            pass

        try:
            cls.math_server_context.__exit__(None, None, None)
        except Exception:
            pass

    async def test_load_tools_with_annotations(self):
        """Test loading tools from a server that uses tool annotations."""
        client = MultiServerMCPClient({"time": self.time_connection})

        tools = await client.get_tools(server_name="time")
        self.assertEqual(len(tools), 1)

        tool = tools[0]
        self.assertEqual(tool.name, "get_time")
        self.assertIn("Get current time", tool.description)

    async def test_load_tools_with_custom_httpx_client_factory(self):
        """Test loading tools with a custom HTTPX client factory."""

        def custom_httpx_client_factory(
            headers: dict[str, str] | None = None,
            timeout: httpx.Timeout | None = None,
            auth: httpx.Auth | None = None,
        ) -> httpx.AsyncClient:
            return httpx.AsyncClient(
                headers=headers,
                timeout=timeout or httpx.Timeout(30.0),
                auth=auth,
                follow_redirects=True,
                limits=httpx.Limits(max_keepalive_connections=5, max_connections=10),
            )

        status_connection_with_factory = {
            **self.status_connection,
            "httpx_client_factory": custom_httpx_client_factory,
        }

        client = MultiServerMCPClient({"status": status_connection_with_factory})

        tools = await client.get_tools(server_name="status")
        self.assertEqual(len(tools), 2)

        tool_names = {tool.name for tool in tools}
        self.assertEqual(tool_names, {"get_status", "get_uptime"})

        status_tool = next(tool for tool in tools if tool.name == "get_status")
        result = await status_tool()
        self.assertEqual(result["response"], "Server is running")

    async def test_load_tools_from_specific_server_no_namespacing(self):
        """Test loading tools from a specific server (no namespacing applied)."""
        client = MultiServerMCPClient(
            {
                "time": self.time_connection,
                "status": self.status_connection,
                "math": self.math_connection,
            }
        )

        # Load tools from specific server - no namespacing
        math_tools = await client.get_tools(server_name="math")
        self.assertEqual(len(math_tools), 2)

        tool_names = {tool.name for tool in math_tools}
        self.assertEqual(tool_names, {"add_numbers", "multiply_numbers"})

        add_tool = next(tool for tool in math_tools if tool.name == "add_numbers")
        result = await add_tool(a=5, b=3)
        self.assertEqual(result["response"], "8")

    async def test_load_tools_from_all_servers_with_namespacing(self):
        """Test loading tools from all servers (namespacing applied)."""
        client = MultiServerMCPClient(
            {
                "time": self.time_connection,
                "status": self.status_connection,
                "math": self.math_connection,
            }
        )

        # Load tools from all servers - namespacing applied
        all_tools = await client.get_tools()
        self.assertEqual(len(all_tools), 5)  # 1 + 2 + 2 tools

        tool_names = {tool.name for tool in all_tools}
        expected_names = {
            "time_get_time",
            "status_get_status",
            "status_get_uptime",
            "math_add_numbers",
            "math_multiply_numbers",
        }
        self.assertEqual(tool_names, expected_names)

        # Test calling namespaced tools
        namespaced_add_tool = next(
            tool for tool in all_tools if tool.name == "math_add_numbers"
        )
        result = await namespaced_add_tool(a=10, b=7)
        self.assertEqual(result["response"], "17")

        namespaced_time_tool = next(
            tool for tool in all_tools if tool.name == "time_get_time"
        )
        result = await namespaced_time_tool()
        self.assertEqual(result["response"], "5:20:00 PM EST")



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import AsyncMock
from unittest.mock import MagicMock

from mcp.types import CallToolResult
from mcp.types import EmbeddedResource
from mcp.types import ImageContent
from mcp.types import TextContent
from mcp.types import TextResourceContents
from mcp.types import Tool as MCPTool

from synalinks.src import testing
from synalinks.src.modules.core.tool import Tool
from synalinks.src.utils.mcp.tools import ToolException
from synalinks.src.utils.mcp.tools import _convert_call_tool_result
from synalinks.src.utils.mcp.tools import convert_mcp_tool_to_synalinks_tool
from synalinks.src.utils.mcp.tools import load_mcp_tools


class MCPToolsTest(testing.TestCase):
    def test_convert_empty_text_content(self):
        result = CallToolResult(
            content=[],
            isError=False,
        )

        tool_message = _convert_call_tool_result(result)

        self.assertEqual(tool_message["response"], "")

    def test_convert_single_text_content(self):
        result = CallToolResult(
            content=[TextContent(type="text", text="test result")],
            isError=False,
        )

        tool_message = _convert_call_tool_result(result)

        self.assertEqual(tool_message["response"], "test result")

    def test_convert_multiple_text_contents(self):
        result = CallToolResult(
            content=[
                TextContent(type="text", text="result 1"),
                TextContent(type="text", text="result 2"),
            ],
            isError=False,
        )

        tool_message = _convert_call_tool_result(result)

        self.assertEqual(tool_message["response"], ["result 1", "result 2"])

    def test_convert_with_non_text_content(self):
        image_content = ImageContent(
            type="image", mimeType="image/png", data="base64data"
        )
        resource_content = EmbeddedResource(
            type="resource",
            resource=TextResourceContents(
                uri="resource://test", mimeType="text/plain", text="hi"
            ),
        )

        result = CallToolResult(
            content=[
                TextContent(type="text", text="text result"),
                image_content,
                resource_content,
            ],
            isError=False,
        )

        tool_message = _convert_call_tool_result(result)

        self.assertEqual(tool_message["response"], "text result")
        self.assertNotIn("artifact", tool_message)

    def test_convert_with_error(self):
        result = CallToolResult(
            content=[TextContent(type="text", text="error message")],
            isError=True,
        )

        with self.assertRaises(ToolException) as exc_info:
            _convert_call_tool_result(result)

        self.assertEqual(str(exc_info.exception), "error message")

    async def test_convert_mcp_tool_to_synalinks_tool(self):
        tool_input_schema = {
            "properties": {
                "param1": {"title": "Param1", "type": "string"},
                "param2": {"title": "Param2", "type": "integer"},
            },
            "required": ["param1", "param2"],
            "title": "ToolSchema",
            "type": "object",
        }

        session = AsyncMock()
        session.call_tool.return_value = CallToolResult(
            content=[TextContent(type="text", text="tool result")],
            isError=False,
        )

        mcp_tool = MCPTool(
            name="test_tool",
            description="Test tool description",
            inputSchema=tool_input_schema,
        )

        synalinks_tool = convert_mcp_tool_to_synalinks_tool(session, mcp_tool)

        self.assertIsInstance(synalinks_tool, Tool)
        self.assertEqual(synalinks_tool.name, "test_tool")
        self.assertStartsWith(synalinks_tool.description, "Test tool description")

        result = await synalinks_tool(param1="test", param2=42)

        session.call_tool.assert_called_once_with(
            "test_tool", {"param1": "test", "param2": 42}
        )

        # New Tool module returns JsonDataModel, use get_json() to access data
        self.assertEqual(result.get_json()["response"], "tool result")

    async def test_load_mcp_tools(self):
        tool_input_schema = {
            "properties": {
                "param1": {"title": "Param1", "type": "string"},
                "param2": {"title": "Param2", "type": "integer"},
            },
            "required": ["param1", "param2"],
            "title": "ToolSchema",
            "type": "object",
        }

        session = AsyncMock()
        mcp_tools = [
            MCPTool(
                name="tool1",
                description="Tool 1 description",
                inputSchema=tool_input_schema,
            ),
            MCPTool(
                name="tool2",
                description="Tool 2 description",
                inputSchema=tool_input_schema,
            ),
        ]
        session.list_tools.return_value = MagicMock(tools=mcp_tools, nextCursor=None)

        async def mock_call_tool(tool_name, arguments):
            if tool_name == "tool1":
                return CallToolResult(
                    content=[
                        TextContent(type="text", text=f"tool1 result with {arguments}")
                    ],
                    isError=False,
                )
            elif tool_name == "tool2":
                return CallToolResult(
                    content=[
                        TextContent(type="text", text=f"tool2 result with {arguments}")
                    ],
                    isError=False,
                )
            else:
                raise ValueError(f"Unknown tool: {tool_name}")

        session.call_tool.side_effect = mock_call_tool

        toolkit = await load_mcp_tools(session)

        self.assertEqual(len(toolkit), 2)
        self.assertTrue(all(isinstance(tool, Tool) for tool in toolkit))
        self.assertEqual(toolkit[0].name, "tool1")
        self.assertStartsWith(toolkit[0].description, "Tool 1 description")
        self.assertEqual(toolkit[1].name, "tool2")
        self.assertStartsWith(toolkit[1].description, "Tool 2 description")

        result1 = await toolkit[0](param1="test1", param2=1)
        self.assertEqual(
            result1.get_json()["response"],
            "tool1 result with {'param1': 'test1', 'param2': 1}",
        )

        result2 = await toolkit[1](param1="test2", param2=2)
        self.assertEqual(
            result2.get_json()["response"],
            "tool2 result with {'param1': 'test2', 'param2': 2}",
        )

    async def test_load_mcp_tools_with_namespace(self):
        tool_input_schema = {
            "properties": {
                "param1": {"title": "Param1", "type": "string"},
                "param2": {"title": "Param2", "type": "integer"},
            },
            "required": ["param1", "param2"],
            "title": "ToolSchema",
            "type": "object",
        }

        namespace = "testing"

        session = AsyncMock()
        mcp_tools = [
            MCPTool(
                name="tool1",
                description="Tool 1 description",
                inputSchema=tool_input_schema,
            ),
            MCPTool(
                name="tool2",
                description="Tool 2 description",
                inputSchema=tool_input_schema,
            ),
        ]
        session.list_tools.return_value = MagicMock(tools=mcp_tools, nextCursor=None)

        async def mock_call_tool(tool_name, arguments):
            if tool_name == "tool1":
                return CallToolResult(
                    content=[
                        TextContent(type="text", text=f"tool1 result with {arguments}")
                    ],
                    isError=False,
                )
            elif tool_name == "tool2":
                return CallToolResult(
                    content=[
                        TextContent(type="text", text=f"tool2 result with {arguments}")
                    ],
                    isError=False,
                )
            else:
                raise ValueError(f"Unknown tool: {tool_name}")

        session.call_tool.side_effect = mock_call_tool

        toolkit = await load_mcp_tools(session, namespace=namespace)

        self.assertEqual(len(toolkit), 2)
        self.assertTrue(all(isinstance(tool, Tool) for tool in toolkit))
        self.assertEqual(toolkit[0].name, f"{namespace}_tool1")
        self.assertStartsWith(toolkit[0].description, "Tool 1 description")
        self.assertEqual(toolkit[1].name, f"{namespace}_tool2")
        self.assertStartsWith(toolkit[1].description, "Tool 2 description")

        result1 = await toolkit[0](param1="test1", param2=1)
        self.assertEqual(
            result1.get_json()["response"],
            "tool1 result with {'param1': 'test1', 'param2': 1}",
        )

        result2 = await toolkit[1](param1="test2", param2=2)
        self.assertEqual(
            result2.get_json()["response"],
            "tool2 result with {'param1': 'test2', 'param2': 2}",
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import synalinks
from synalinks.src import testing
from synalinks.src.modules.core.tool import Tool
from synalinks.src.modules.core.tool import json_schema_type
from synalinks.src.modules.module import Module


class ToolModuleTest(testing.TestCase):
    """Tests for the Tool module."""

    def test_inheritance(self):
        """Test that Tool inherits from Module."""
        self.assertTrue(issubclass(Tool, Module))

    def test_init_with_valid_function(self):
        """Test initialization with a valid async function."""

        @synalinks.saving.register_synalinks_serializable()
        async def my_tool(arg1: str):
            """A test tool.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        tool = Tool(func=my_tool)
        self.assertEqual(tool.name, "my_tool")
        self.assertEqual(tool.description, "A test tool.")
        self.assertFalse(tool.trainable)

    def test_init_with_custom_name(self):
        """Test initialization with custom name and description."""

        @synalinks.saving.register_synalinks_serializable()
        async def my_tool(arg1: str):
            """A test tool.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        tool = Tool(func=my_tool, name="custom_name", description="Custom description")
        self.assertEqual(tool.name, "custom_name")
        self.assertEqual(tool.description, "Custom description")

    def test_init_with_sync_function_raises(self):
        """Test that sync functions raise TypeError."""

        def sync_func(arg1: str):
            """A sync function.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        with self.assertRaises(TypeError):
            Tool(func=sync_func)

    def test_init_without_docstring_raises(self):
        """Test that functions without docstrings raise ValueError."""

        async def no_docstring(arg1: str):
            return {"result": arg1}

        with self.assertRaises(ValueError):
            Tool(func=no_docstring)

    async def test_call_executes_function(self):
        """Test that call executes the wrapped function."""

        @synalinks.saving.register_synalinks_serializable()
        async def add_numbers(a: int, b: int):
            """Add two numbers.

            Args:
                a (int): First number.
                b (int): Second number.
            """
            return {"sum": a + b}

        tool = Tool(func=add_numbers)
        result = await tool(a=5, b=3)
        self.assertEqual(result.get_json(), {"sum": 8})

    async def test_call_with_training_flag(self):
        """Test that call works with training flag."""

        @synalinks.saving.register_synalinks_serializable()
        async def simple_tool(value: str):
            """A simple tool.

            Args:
                value (str): A string value.
            """
            return {"value": value}

        tool = Tool(func=simple_tool)
        result = await tool(training=True, value="test")
        self.assertEqual(result.get_json(), {"value": "test"})

    def test_get_tool_schema(self):
        """Test get_tool_schema returns correct schema."""

        @synalinks.saving.register_synalinks_serializable()
        async def test_func(required_arg: str, optional_arg: int = 10):
            """Test function with required and optional args.

            Args:
                required_arg (str): A required string argument.
                optional_arg (int): An optional integer argument.
            """
            return {"result": required_arg}

        tool = Tool(func=test_func)
        schema = tool.get_tool_schema()

        self.assertEqual(schema["type"], "object")
        self.assertIn("required_arg", schema["properties"])
        self.assertIn("optional_arg", schema["properties"])
        self.assertIn("required_arg", schema["required"])
        self.assertNotIn("optional_arg", schema["required"])
        self.assertEqual(schema["properties"]["required_arg"]["type"], "string")
        self.assertEqual(schema["properties"]["optional_arg"]["type"], "integer")
        self.assertEqual(schema["properties"]["optional_arg"]["default"], 10)

    def test_get_config(self):
        """Test get_config returns serializable config."""

        @synalinks.saving.register_synalinks_serializable()
        async def my_tool(arg1: str):
            """A test tool.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        tool = Tool(func=my_tool, name="test_tool", description="Test description")
        config = tool.get_config()

        self.assertIn("func", config)
        self.assertEqual(config["name"], "test_tool")
        self.assertEqual(config["description"], "Test description")
        self.assertFalse(config["trainable"])

    async def test_from_config(self):
        """Test from_config reconstructs the tool."""

        @synalinks.saving.register_synalinks_serializable()
        async def my_tool(arg1: str):
            """A test tool.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        original_tool = Tool(func=my_tool)
        config = original_tool.get_config()

        restored_tool = Tool.from_config(config)
        self.assertEqual(restored_tool.name, original_tool.name)
        self.assertEqual(restored_tool.description, original_tool.description)

        # Test that the restored tool works
        result = await restored_tool(arg1="test")
        self.assertEqual(result.get_json(), {"result": "test"})


class JsonSchemaTypeTest(testing.TestCase):
    """Tests for json_schema_type function."""

    def test_basic_types(self):
        """Test conversion of basic Python types."""
        self.assertEqual(json_schema_type(int), "integer")
        self.assertEqual(json_schema_type(float), "number")
        self.assertEqual(json_schema_type(bool), "boolean")
        self.assertEqual(json_schema_type(str), "string")
        self.assertEqual(json_schema_type(type(None)), "null")

    def test_list_type(self):
        """Test conversion of list types."""
        from typing import List

        result = json_schema_type(List[str])
        self.assertEqual(result["type"], "array")
        self.assertEqual(result["items"]["type"], "string")

    def test_dict_type(self):
        """Test conversion of dict types."""
        from typing import Dict

        result = json_schema_type(Dict[str, int])
        self.assertEqual(result["type"], "object")
        self.assertEqual(result["additionalProperties"]["type"], "integer")

    def test_optional_type(self):
        """Test conversion of Optional types."""
        from typing import Optional

        result = json_schema_type(Optional[str])
        self.assertEqual(result, "string")

    def test_unsupported_type_raises(self):
        """Test that unsupported types raise ValueError."""

        class CustomClass:
            pass

        with self.assertRaises(ValueError):
            json_schema_type(CustomClass)


class ToolOutputSchemaTest(testing.TestCase):
    """Tests for Tool output schema generation."""

    def test_output_schema_no_return_type(self):
        """Test output schema when no return type is specified."""

        @synalinks.saving.register_synalinks_serializable()
        async def no_return_type(arg1: str):
            """A tool with no return type.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        tool = Tool(func=no_return_type)
        schema = tool._build_output_schema()

        self.assertEqual(schema["type"], "object")
        self.assertEqual(schema["title"], "no_return_type_output")
        self.assertTrue(schema["additionalProperties"])

    def test_output_schema_dict_return_type(self):
        """Test output schema with Dict[str, str] return type."""
        from typing import Dict

        @synalinks.saving.register_synalinks_serializable()
        async def dict_return(arg1: str) -> Dict[str, str]:
            """A tool with dict return type.

            Args:
                arg1 (str): The first argument.
            """
            return {"result": arg1}

        tool = Tool(func=dict_return)
        schema = tool._build_output_schema()

        self.assertEqual(schema["type"], "object")
        self.assertEqual(schema["title"], "dict_return_output")
        self.assertEqual(schema["additionalProperties"]["type"], "string")

    def test_output_schema_typed_dict_return(self):
        """Test output schema with TypedDict return type."""
        from typing import TypedDict

        class CalculationResult(TypedDict):
            result: int
            message: str

        @synalinks.saving.register_synalinks_serializable()
        async def typed_dict_return(a: int, b: int) -> CalculationResult:
            """A tool with TypedDict return type.

            Args:
                a (int): First number.
                b (int): Second number.
            """
            return {"result": a + b, "message": "Success"}

        tool = Tool(func=typed_dict_return)
        schema = tool._build_output_schema()

        self.assertEqual(schema["type"], "object")
        self.assertIn("properties", schema)
        self.assertIn("result", schema["properties"])
        self.assertIn("message", schema["properties"])
        self.assertEqual(schema["properties"]["result"]["type"], "integer")
        self.assertEqual(schema["properties"]["message"]["type"], "string")
        self.assertFalse(schema["additionalProperties"])

    def test_get_input_schema(self):
        """Test that get_input_schema returns correct input schema."""

        @synalinks.saving.register_synalinks_serializable()
        async def my_tool(name: str, count: int = 5):
            """A tool with multiple parameters.

            Args:
                name (str): The name parameter.
                count (int): The count parameter.
            """
            return {"name": name, "count": count}

        tool = Tool(func=my_tool)
        schema = tool.get_input_schema()

        self.assertEqual(schema["type"], "object")
        self.assertIn("name", schema["properties"])
        self.assertIn("count", schema["properties"])
        self.assertIn("name", schema["required"])
        self.assertNotIn("count", schema["required"])
        self.assertEqual(schema["properties"]["name"]["type"], "string")
        self.assertEqual(schema["properties"]["count"]["type"], "integer")



# Modified from: keras/src/layers/layer.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import collections
import inspect
import uuid
import warnings
from functools import wraps

from synalinks.src import backend
from synalinks.src import initializers
from synalinks.src import tree
from synalinks.src import utils
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import is_trainable
from synalinks.src.backend.common import global_state
from synalinks.src.backend.common.name_scope import current_path
from synalinks.src.hooks.hook_list import HookList
from synalinks.src.metrics import Metric
from synalinks.src.ops.operation import Operation
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable
from synalinks.src.utils import python_utils
from synalinks.src.utils import tracking
from synalinks.src.utils.async_utils import run_maybe_nested
from synalinks.src.utils.naming import auto_name

if backend.backend() == "pydantic":
    from synalinks.src.backend.pydantic.module import PydanticModule as BackendModule
else:
    raise RuntimeError(
        f"Backend '{backend.backend()}' must implement a module mixin class."
    )


@synalinks_export(["synalinks.Module", "synalinks.modules.Module"])
class Module(BackendModule, Operation, SynalinksSaveable):
    """This is the class from which all modules inherit.

    A module is a callable object that takes as input one or more data models and
    that outputs one or more data models. It involves *computation*, defined
    in the `call()` method, and a *state* (the variables). State can be
    created:

    * in `__init__()`, for instance via `self.add_variable()`;
    * in the optional `build()` method, which is invoked by the first
      `__call__()` to the module, and supplies the schema(s) of the input(s),
      which may not have been known at initialization time.

    Modules are recursively composable: If you assign a Module instance as an
    attribute of another Module, the outer Module will start tracking the variables
    created by the inner module. Nested modules should be instantiated in the
    `__init__()` method or `build()` method.

    Users will just instantiate a module and then treat it as a callable.

    Args:
        trainable (bool): Boolean, whether the module's variables should be trainable.
        name (str): String name of the module.

    We recommend that descendants of `Module` implement the following methods:

    * `__init__()`: Defines custom modules attributes, and creates module variables
        that do not depend on input schemas, using `add_variable()`,
        or other state.
    * `build(self, input_schema)`: This method can be used to create variables that
        depend on the schemas(s) of the input(s), using `add_variable()`, or other
        state. `__call__()` will automatically build the module
        (if it has not been built yet) by calling `build()`.
    * `call(self, *args, **kwargs)`: Called in `__call__` after making
        sure `build()` has been called. `call()` performs the logic of applying
        the module to the input arguments.
        Two reserved keyword arguments you can optionally use in `call()` are:
            1. `training` (boolean, whether the call is in inference mode or
                training mode).
        A typical signature for this method is `call(self, inputs)`, and user
        could optionally add `training` if the module need it.
    * `get_config(self)`: Returns a dictionary containing the configuration
        used to initialize this module. If the keys differ from the arguments
        in `__init__()`, then override `from_config(self)` as well.
        This method is used when saving
        the module or a program that contains this module.
    """

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)  # , *args, **kwargs)

        # Wrap the user-provided `build` method in the `build_wrapper`
        # to add name scope support and serialization support.
        original_build_method = obj.build

        @wraps(original_build_method)
        async def build_wrapper(*args, **kwargs):
            with obj._open_name_scope():
                obj._path = current_path()
                await original_build_method(*args, **kwargs)
            # Record build config.
            signature = inspect.signature(original_build_method)
            obj._build_schemas_dict = signature.bind(*args, **kwargs).arguments
            # Set built, post build actions, and lock state.
            obj.built = True
            obj._post_build()
            obj._lock_state()

        obj.build = build_wrapper
        return obj

    def __init__(
        self,
        *,
        trainable=True,
        name=None,
        description=None,
        hooks=None,
        **kwargs,
    ):
        BackendModule.__init__(self)
        self._lock = False
        Operation.__init__(self, name=name, description=description)
        if kwargs:
            raise ValueError(
                "Unrecognized keyword arguments "
                f"passed to {self.__class__.__name__}: {kwargs}"
            )
        self._path = None  # Will be determined in `build_wrapper`
        self.built = False
        self._input_spec = None
        self._called = False

        self._trainable = trainable
        self._rewards = []
        self._reward_ids = set()
        self._rewards_override = []

        self._call_signature = inspect.signature(self.call)
        call_signature_parameters = [
            p.name for p in self._call_signature.parameters.values()
        ]
        self._call_has_training_arg = "training" in call_signature_parameters
        # Whether to automatically convert inputs to `call()`.
        self._convert_input_args = True
        # Whether to allow non-json object as positional arguments in `call()`.
        self._allow_non_json_data_model_positional_args = False
        # Dict of schemas that were used to call `build()`.
        self._build_schemas_dict = None
        # Parent path
        self._parent_path = None
        self._hooks = HookList(
            hooks=hooks,
            module=self,
        )
        self._initialize_tracker()

    @tracking.no_automatic_dependency_tracking
    def _initialize_tracker(self):
        if hasattr(self, "_tracker"):
            return

        trainable_variables = []
        non_trainable_variables = []
        modules = []
        metrics = []
        self._tracker = tracking.Tracker(
            {
                "trainable_variables": (
                    lambda x: isinstance(x, backend.Variable) and x.trainable,
                    trainable_variables,
                ),
                "non_trainable_variables": (
                    lambda x: isinstance(x, backend.Variable) and not x.trainable,
                    non_trainable_variables,
                ),
                "metrics": (lambda x: isinstance(x, Metric), metrics),
                "modules": (
                    lambda x: isinstance(x, Module) and not isinstance(x, Metric),
                    modules,
                ),
            },
            exclusions={"non_trainable_variables": ["trainable_variables"]},
        )

        self._trainable_variables = trainable_variables
        self._non_trainable_variables = non_trainable_variables
        self._modules = modules
        self._metrics = metrics

    @property
    def path(self):
        """The path of the module.

        If the module has not been built yet, it will be `None`.
        """
        return self._path

    @property
    def input_spec(self):
        return self._input_spec

    @input_spec.setter
    def input_spec(self, value):
        self._input_spec = value

    @classmethod
    def get_arity(cls):
        # Inspect the call method to get the number of parameters
        sig = inspect.signature(cls.call)
        # Exclude 'self' and 'training' from the parameter count
        return len(
            [
                param
                for param in sig.parameters.values()
                if param.name not in ["self", "training"]
            ]
        )

    @python_utils.default
    async def build(self, *args, **kwargs):
        self._check_super_called()
        if utils.is_default(self.build) and might_have_unbuilt_state(self):
            try:
                if isinstance(args, tuple):
                    args = list(args)
                if len(args) == 1 or backend.is_data_model(args[0]):
                    args[0] = args[0].to_symbolic_data_model()
                else:
                    args = tree.map_structure(
                        lambda x: (
                            x.to_symbolic_data_model() if backend.is_data_model(x) else x
                        ),
                        args,
                    )
                await self.__call__(*args, **kwargs)
            except Exception as e:
                warnings.warn(
                    f"`build()` was called on module '{self.name}', however "
                    "the module does not have a `build()` method implemented "
                    "and it looks like it has unbuilt state. This will cause "
                    "the module to be marked as built, despite not being "
                    "actually built, which may cause failures down the line. "
                    "Make sure to implement a proper `build()` method."
                    f"Exception encountered: ''{e}''"
                )
        self.built = True

    def _lock_state(self):
        """Prevent further state updates, called automatically in `build()`."""
        if not self._tracker.locked:
            self._tracker.lock(
                msg=(
                    "You cannot add new elements of state "
                    "(variables or sub-modules) "
                    "to a module that is already built. All state "
                    "must be created in the `__init__()` method or "
                    "in the `build()` method."
                )
            )

    def get_build_config(self):
        """Returns a dictionary with the modules's input schema.

        This method returns a config dict that can be used by
        `build_from_config(config)` to create all states (e.g. Variables and
        Lookup tables) needed by the module.

        By default, the config only contains the input schema that the module
        was built with. If you're writing a custom module that creates state in
        an unusual way, you should override this method to make sure this state
        is already created when Synalinks attempts to load its value upon model
        loading.

        Returns:
            (dict): A dict containing the input schema associated with the module.
        """
        if self._build_schemas_dict is not None:
            if len(self._build_schemas_dict) == 1:
                return {
                    "input_schema": tuple(self._build_schemas_dict.values())[0],
                }
            else:
                return {"schemas_dict": self._build_schemas_dict}

    def build_from_config(self, config):
        """Builds the module's states with the supplied config dict.

        By default, this method calls the `build()` method,
        which creates variables based on the module's input schema in the supplied
        config. If your config contains other information needed to load the
        module's state, you should override this method.

        Args:
            config (dict): Dict containing the input schema associated with this module.
        """
        if config:
            if "input_schema" in config:
                run_maybe_nested(
                    self.build(backend.SymbolicDataModel(schema=config["input_schema"]))
                )
            elif "schemas_dict" in config:
                symbolic_inputs = {}
                for key, schema in config["schemas_dict"].items():
                    symbolic_inputs[key] = backend.SymbolicDataModel(schema=schema)
                run_maybe_nested(self.build(**symbolic_inputs))
            self.built = True

    def _obj_type(self):
        return "Module"

    @property
    def metrics(self):
        """List of all metrics."""
        metrics = list(self._metrics)
        for module in self._modules:
            metrics.extend(module.metrics)
        return metrics

    @property
    def metrics_variables(self):
        """List of all metric variables."""
        vars = []
        for metric in self.metrics:
            vars.extend(metric.variables)
        return vars

    def _get_own_rewards(self):
        if backend.in_stateless_scope():
            rewards = []
            scope = backend.get_stateless_scope()
            for reward in scope.rewards:
                if id(reward) in self._reward_ids:
                    rewards.append(reward)
            return rewards
        else:
            return self._rewards[:]

    @property
    def rewards(self):
        """List of scalar rewards from `add_reward` and submodules."""
        if self._rewards_override:
            return self._rewards_override
        rewards = self._get_own_rewards()
        for module in self._flatten_modules(include_self=False):
            rewards.extend(module._get_own_rewards())
        return rewards

    def _clear_rewards(self):
        if backend.in_stateless_scope():
            scope = backend.get_stateless_scope()
            if scope.collect_rewards:
                for x in scope.rewards:
                    if id(x) in self._reward_ids:
                        scope.rewards.remove(x)
        self._rewards.clear()
        self._reward_ids.clear()
        for module in self._modules:
            module._clear_rewards()

    def add_variable(
        self,
        initializer=None,
        data_model=None,
        trainable=True,
        name=None,
        description=None,
    ):
        """Add a variable to the module

        Args:
            initializer (dict | Initializer): Initializer object to use to
                populate the initial variable value. Can be a JSON dict containing the
                initial value. If unspecified, defaults to `initializers.Empty`.
            data_model (DataModel): The DataModel used to infer the schema
                and default value.
            trainable (bool): Boolean, whether the variable should be trainable via
                optimization or whether its updates are managed manually. Defaults
                to `True`.
            name (string): String name of the variable. Useful for debugging purposes.
            description (string): String description of the variable. Used by the
                optimizers to infer the role of the variable. Required if the data
                model do not have a docstring.

        Returns:
            (Variable): The created variable
        """
        self._check_super_called()
        if initializer is None:
            initializer = initializers.Empty(data_model=data_model)
        trainable = trainable and is_trainable(data_model)
        with backend.name_scope(self.name, caller=self):
            variable = backend.Variable(
                initializer=initializer,
                data_model=data_model,
                trainable=trainable,
                name=name,
                description=description,
            )
        self._track_variable(variable)
        return variable

    @property
    def trainable(self):
        """Settable boolean, whether this module should be trainable or not."""
        return self._trainable

    @trainable.setter
    def trainable(self, value):
        """Sets trainable attribute for the module and its submodules.

        When this value is changed during training (e.g. with a
        `Callback`) you need to call the parent
        `Program.make_train_function` with `force=True` in order to
        recompile the training graph.

        Args:
            value (bool): Boolean with the desired state for the module's trainable
                attribute.
        """
        value = bool(value)
        self._trainable = value
        for v in self._trainable_variables:
            v.trainable = value
        for module in self._modules:
            module.trainable = value

    def add_hook(self, hook):
        self._hooks.add_hook(hook)

    @property
    def variables(self):
        """List of all module state.

        Note that metrics variables are not included here, use
        `metrics_variables` to visit all the metric variables.

        Returns:
            (list): The list of the variables.
        """
        # Return all `Variables` associate with the module including metrics
        # and random seeds. Also deduplicate them.
        variables = []
        seen_ids = set()
        for v in self._trainable_variables + self._non_trainable_variables:
            if id(v) not in seen_ids:
                variables.append(v)
                seen_ids.add(id(v))
        for module in self._modules:
            for v in module.variables:
                if id(v) not in seen_ids:
                    variables.append(v)
                    seen_ids.add(id(v))
        return variables

    @property
    def trainable_variables(self):
        """List of all trainable module state.

        Returns:
            (list): The list of trainable variables.
        """
        if not self.trainable:
            return []
        return [v for v in self.variables if v.trainable]

    @property
    def non_trainable_variables(self):
        """List of all non-trainable module state.

        Returns:
            (list): The list of non-trainable variables.
        """
        if not self.trainable:
            return self.variables
        return [v for v in self.variables if not v.trainable]

    def get_variable(self, name=None, index=None):
        """Retrieves a variable based on either its name (unique) or index.

        If `name` and `index` are both provided, `index` will take precedence.
        Indices are based on order of instantiation.

        Args:
            name (string): The name of the variable.
            index (int): The index of the variable.

        Returns:
            (Variable): The returned variable.
        """
        if index is not None and name is not None:
            raise ValueError(
                "Provide only a variable name or a variable index. Received: "
                f"index={index}, name={name}."
            )
        if index is not None:
            if len(self.variables) <= index:
                raise ValueError(
                    f"Was asked to retrieve variable at index {index}"
                    f" but module only has {len(self.modules)}"
                    " variables."
                )
            else:
                return self.variables[index]

        if name is not None:
            for variable in self.variables:
                if variable.name == name:
                    return variable
            raise ValueError(
                f"No such variable: {name}. Existing variables are: "
                f"{list(variable.name for variable in self.variables)}."
            )
        raise ValueError(
            "Provide either a variable name or variable index at `get_variable`."
        )

    async def __call__(self, *args, **kwargs):
        call_id = str(uuid.uuid4())

        self._check_super_called()
        self._called = True

        call_context = self._get_call_context()

        parent_call_id = call_context.call_id if call_context.call_id else None

        call_context.call_id = call_id

        if self._hooks:
            self._hooks.on_call_begin(
                call_id=call_id,
                parent_call_id=parent_call_id,
                inputs=args,
                kwargs=kwargs,
            )

        #####################################
        # 0. Convert tuple inputs to list for convenience
        if isinstance(args, tuple):
            args = list(args)

        #####################################
        # 1. Convert any DataModel positional arguments to JsonDataModel
        # This operation is performed to make the computation backend independent
        # and the making the data models dynamically modifiable

        # Used to avoid expensive `tree` operations in the most common case.
        if len(args) == 1 and backend.is_data_model(args[0]):
            args[0] = args[0].to_json_data_model(name="inputs_" + self.name)
        else:
            args = self._maybe_convert_inputs(args)

        ##########################################################
        # 2. Enforce that only JsonDataModels or SymbolicDataModel
        # can be passed positionally.
        if not self._allow_non_json_data_model_positional_args:
            for arg in tree.flatten(args):
                if not is_json_data_model_or_symbolic_data_model(arg) and arg is not None:
                    raise ValueError(
                        "Only input JsonDataModel, DataModel or SymbolicDataModel"
                        " may be passed as positional arguments. The following argument "
                        f"value should be passed as a keyword argument: {arg} "
                        f"(of type {type(arg)})"
                    )

        # Caches info about `call()` signature, args, kwargs.
        call_spec = CallSpec(self._call_signature, args, kwargs)

        ############################################
        # 3. Check input spec for 1st positional arg.
        # TODO: consider extending this to all args and kwargs.
        self._assert_input_compatibility(call_spec.first_arg)

        ################
        # 4. Call build
        with self._open_name_scope():
            await self._maybe_build(call_spec)

        ##########################
        # 5. Infer training value
        # Training phase for `Module.call` is set via (in order of priority):
        # (1) The `training` argument passed to this `Module.call`, if not None
        # (2) The training argument of an outer `Module.call`.
        # (4) Any non-None default value for `training` in the call signature
        # (5) False (treating the module as if it's in inference)

        # This is the value explicitly passed by the user
        training = call_spec.user_arguments_dict.get("training", None)
        if training is None:
            # Wasn't passed explicitly: use context value
            training = call_context.training
            if training is None:
                # Get signature default value
                training = call_spec.arguments_dict.get("training", None)
        call_context.training = training
        if self._call_has_training_arg and training is not None:
            # Only populate arg if it has a concrete value
            kwargs["training"] = training

        ####################
        # 6. Call the module.
        try:
            with self._open_name_scope():
                outputs = await super().__call__(*args, **kwargs)

            if not self.built:
                self.built = True
        except Exception as e:
            if self._hooks:
                self._hooks.on_call_end(
                    call_id=call_id,
                    exception=str(e),
                )
            raise e
        finally:
            # Destroy call context if we created it
            self._maybe_reset_call_context()
        if self._hooks:
            self._hooks.on_call_end(
                call_id=call_id,
                parent_call_id=parent_call_id,
                outputs=outputs,
            )
        return outputs

    async def call(self, *args, **kwargs):
        raise self._not_implemented_error(self.call)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            f"name={self.name}, description='{self.description}', built={self.built}>"
        )

    def __str__(self):
        return self.__repr__()

    def __setattr__(self, name, value):
        # Track Variables, Modules, Metrics.
        name, value = self._setattr_hook(name, value)
        if name != "_tracker":
            if not hasattr(self, "_tracker"):
                self._initialize_tracker()
            value = self._tracker.track(value)
        return super().__setattr__(name, value)

    def __delattr__(self, name):
        obj = getattr(self, name)
        if isinstance(obj, backend.Variable):
            import gc

            # It will take a short amount of time for the corresponding buffer
            # to be actually removed from the device.
            # https://stackoverflow.com/a/74631949
            self._untrack_variable(obj)
            super().__delattr__(name)
            gc.collect()
        else:
            super().__delattr__(name)

    def _check_super_called(self):
        if getattr(self, "_lock", True):
            raise RuntimeError(
                f"In module '{self.__class__.__name__}', you forgot to call "
                "`super().__init__()` as the first statement "
                "in the `__init__()` method. Go add it!"
            )

    def _assert_input_compatibility(self, first_arg):
        # TODO perform check using schemas
        pass

    def _maybe_convert_inputs(self, inputs):
        counter = {"i": 0}

        def convert_fn(x):
            if backend.is_data_model(x):
                if counter["i"] > 0:
                    result = x.to_json_data_model(
                        name=f"{self.name}_inputs_{counter['i']}"
                    )
                else:
                    result = x.to_json_data_model(name=f"{self.name}_inputs")
                counter["i"] += 1
                return result
            return x

        return tree.map_structure(
            convert_fn,
            inputs,
        )
        return inputs

    def _get_call_context(self):
        """Returns currently active `CallContext`."""
        module_call_ctx = global_state.get_global_attribute("current_call_ctx")
        if module_call_ctx is None:
            # Enter new call context.
            module_call_ctx = CallContext(entry_module=self)
            global_state.set_global_attribute("current_call_ctx", module_call_ctx)
            self._clear_rewards()
        return module_call_ctx

    def _maybe_reset_call_context(self):
        module_call_ctx = global_state.get_global_attribute("current_call_ctx")
        if module_call_ctx is None or module_call_ctx.entry_module == self:
            global_state.set_global_attribute("current_call_ctx", None)

    def _flatten_modules(self, include_self=True, recursive=True):
        modules = []
        if include_self:
            modules.append(self)
        seen_object_ids = set()
        deque = collections.deque(self._modules)
        while deque:
            module = deque.popleft()
            if id(module) in seen_object_ids:
                continue
            seen_object_ids.add(id(module))
            modules.append(module)
            # Introspect recursively through submodules.
            if recursive:
                deque.extendleft(module._modules)
        return modules

    def _not_implemented_error(self, attr, msg=None):
        if callable(attr):
            attr_name = attr.__name__
            attr_type = "method"
        else:
            attr_name = str(attr)
            attr_type = "attribute"
        msg = " " + msg if msg is not None else ""
        return NotImplementedError(
            f"Module {self.__class__.__name__} does not have a `{attr_name}` "
            f"{attr_type} implemented.{msg}"
        )

    def _track_variable(self, variable):
        if variable.trainable:
            self._tracker.add_to_store("trainable_variables", variable)
        else:
            self._tracker.add_to_store("non_trainable_variables", variable)
        if not self.trainable:
            variable.trainable = False
        self._post_track_variable(variable)

    def _untrack_variable(self, variable):
        previous_lock_state = self._tracker.locked
        self._tracker.unlock()
        self._tracker.untrack(variable)
        if previous_lock_state is True:
            self._tracker.lock()
        self._post_untrack_variable(variable)

    @python_utils.default
    def get_config(self):
        self._check_super_called()
        base_config = super().get_config()
        config = {
            "trainable": self.trainable,
        }
        return {**base_config, **config}

    def _open_name_scope(self):
        if self._parent_path is None:
            self._parent_path = current_path()
        return backend.name_scope(self.name, caller=self)

    async def _maybe_build(self, call_spec):
        if self.built:
            return

        # If the module has a build method, call it with our input schemas.
        if not utils.is_default(self.build):
            if len(call_spec.first_arg) == 1:
                await self.build(call_spec.first_arg[0])
            else:
                await self.build(call_spec.first_arg)
            # Check input spec again (after build, since self.input_spec
            # may have been updated
            self._assert_input_compatibility(call_spec.first_arg)
            return

        # Otherwise, attempt to build the module by calling it on symbolic input.
        if might_have_unbuilt_state(self):
            try:
                if not utils.is_default(self.compute_output_spec):
                    await self.compute_output_spec(**call_spec.arguments_dict)
                else:
                    await backend.compute_output_spec(
                        self.call, **call_spec.arguments_dict
                    )
            except Exception as e:
                if call_spec.eager:
                    # Will let the actual eager call do state-building
                    return
                warnings.warn(
                    f"Module '{self.name}' looks like it has unbuilt state, but "
                    "Synalinks is not able to trace the module `call()` in order to "
                    "build it automatically. Possible causes:\n"
                    "1. The `call()` method of your module may be crashing. Try "
                    "to `__call__()` the module eagerly on some test input "
                    "first to see if it works. "
                    "2. If the `call()` method is correct, then you may need "
                    "to implement the `def build(self, inputs)` or "
                    "`def compute_output_spec(inputs, training=False)` method on "
                    "your module."
                    f"Exception encountered: ''{e}''"
                )
        self.built = True

    def clone(self, name=None):
        clone = self.from_config(self.get_config())
        if name:
            clone.name = name
        else:
            clone.name = auto_name(self.name)
        return clone


def is_json_data_model_or_symbolic_data_model(x, allow_none=False):
    if allow_none and x is None:
        return True
    return backend.is_json_data_model(x) or backend.is_symbolic_data_model(x)


class CallSpec:
    def __init__(self, signature, args, kwargs):
        # `training` is a special kwargs that is always available in
        # a module, if user specifies them in their call without adding to spec,
        # we remove them to be able to bind variables. If User is not using
        # `training` anyway so we can ignore.
        if "training" in kwargs and "training" not in signature.parameters:
            kwargs.pop("training")
            bound_args = signature.bind(*args, **kwargs)
        else:
            bound_args = signature.bind(*args, **kwargs)
        self.user_arguments_dict = {k: v for k, v in bound_args.arguments.items()}
        bound_args.apply_defaults()
        arg_dict = {}
        arg_names = []
        data_arg_dict = {}
        data_args = []
        data_arg_names = []
        nested_data_arg_names = []
        for name, value in bound_args.arguments.items():
            arg_dict[name] = value
            arg_names.append(name)
            if is_json_data_model_or_symbolic_data_model(value):
                data_args.append(value)
                data_arg_names.append(name)
                data_arg_dict[name] = value
            elif tree.is_nested(value) and len(value) > 0:
                flat_values = tree.flatten(value)
                if all(
                    is_json_data_model_or_symbolic_data_model(x, allow_none=True)
                    for x in flat_values
                ):
                    data_args.append(value)
                    data_arg_names.append(name)
                    data_arg_dict[name] = value
                    nested_data_arg_names.append(name)
                elif any(
                    is_json_data_model_or_symbolic_data_model(x) for x in flat_values
                ):
                    raise ValueError(
                        "In a nested call() argument, "
                        "you cannot mix JsonDataModels and non-JsonDataModels. "
                        "Received invalid mixed argument: "
                        f"{name}={value}"
                    )
        self.arguments_dict = arg_dict
        self.argument_names = arg_names
        self.data_arguments_dict = data_arg_dict
        self.data_arguments_names = data_arg_names
        self.nested_data_argument_names = nested_data_arg_names
        self.first_arg = arg_dict[arg_names[0]]
        if all(backend.is_json_data_model(x) for x in self.data_arguments_dict.values()):
            self.eager = True
        else:
            self.eager = False


class CallContext:
    def __init__(self, entry_module, call_id=None, parent_call_id=None):
        self.entry_module = entry_module
        self.training = None
        self.call_id = call_id
        self.parent_call_id = parent_call_id
        self.cost = 0.0


def might_have_unbuilt_state(module):
    return any(not module.built for module in module._modules)



# Modified from: keras/src/models/sequential.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import copy
import inspect
import typing

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import standardize_schema
from synalinks.src.modules import Module
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.core.input_module import InputModule
from synalinks.src.programs import Functional
from synalinks.src.programs import Program
from synalinks.src.saving import serialization_lib
from synalinks.src.utils.async_utils import run_maybe_nested


@synalinks_export(["synalinks.Sequential", "synalinks.programs.Sequential"])
class Sequential(Program):
    """`Sequential` groups a linear stack of modules into a `Program`.

    Examples:

    ```python
    program = synalinks.Sequential(
        name="chain_of_thought",
        description="Useful to answer in a step by step manner."
    )
    program.add(
        synalinks.Input(
                data_program=Query,
            )
    )
    program.add(
        synalinks.Generator(
            data_program=AnswerWithRationale,
            language_program=language_program,
        )
    )

    # Note that you can also omit the initial `Input`.
    # In that case the program doesn't have any variables until the first call
    # to a training/evaluation method (since it isn't yet built):

    program = synalinks.Sequential(
        name="chain_of_thought",
        description="Useful to answer in a step by step manner."
    )
    program.add(
        synalinks.Generator(
            data_program=AnswerWithRationale,
            language_program=language_program,
        )
    )
    # program.variables not created yet

    # Whereas if you specify an `Input`, the program gets built
    # continuously as you are adding modules:

    program = synalinks.Sequential(
        name="chain_of_thought",
        description="Useful to answer in a step by step manner."
    )
    program.add(
        synalinks.Input(
            data_program=Query,
        )
    )
    program.add(
        synalinks.Generator(
            data_program=AnswerWithRationale,
            language_program=language_program,
        )
    )

    # Note that when using the delayed-build pattern (no input specified),
    # the program gets built the first time you call `fit`, `eval`, or `predict`,
    # or the first time you call the program on some input data.

    ```
    """

    def __new__(cls, *args, **kwargs):
        return typing.cast(cls, super().__new__(cls))

    def __init__(self, modules=None, trainable=True, name=None, description=None):
        if description is None:
            raise ValueError(
                "All Sequential programs must have a `description`, "
                "please add it to the constructor arguments"
            )
        super().__init__(trainable=trainable, name=name, description=description)
        self._functional = None
        self._modules = []
        if modules:
            for module in modules:
                self.add(module, rebuild=False)
            run_maybe_nested(self._maybe_rebuild())

    def add(self, module, rebuild=True):
        """Adds a module instance on top of the module stack.

        Args:
            module (Module): Module instance.
            rebuild (bool): If `True` rebuild the program.
        """
        # If we are passed a SymbolicDataModel created by synalinks.Input(), we
        # extract the input module from its synalinks history and use that.
        if hasattr(module, "_synalinks_history"):
            origin_module = module._synalinks_history[0]
            if isinstance(origin_module, InputModule):
                module = origin_module
        if not isinstance(module, Module):
            raise ValueError(
                "Only instances of `synalinks.Module` can be "
                f"added to a Sequential program. Received: {module} "
                f"(of type {type(module)})"
            )
        if not self._is_module_name_unique(module):
            raise ValueError(
                "All modules added to a Sequential program "
                f"should have unique names. Name '{module.name}' is already "
                "the name of a module in this program. Update the `name` argument "
                "to pass a unique name."
            )
        if (
            isinstance(module, InputModule)
            and self._modules
            and isinstance(self._modules[0], InputModule)
        ):
            raise ValueError(
                f"Sequential program '{self.name}' has already been configured "
                f"to use input schema {self._modules[0].input_schema}. You cannot "
                f"add a different Input module to it."
            )

        self._modules.append(module)
        if rebuild:
            run_maybe_nested(self._maybe_rebuild())
        else:
            self.built = False
            self._functional = None

    def pop(self, rebuild=True):
        """Removes the last module in the program.

        Args:
            rebuild (bool): If `True` rebuild the program.
        """
        module = self._modules.pop()
        self.built = False
        self._functional = None
        if rebuild:
            run_maybe_nested(self._maybe_rebuild())
        return module

    async def _maybe_rebuild(self):
        self.built = False
        self._functional = None
        if isinstance(self._modules[0], InputModule) and len(self._modules) > 1:
            input_schema = self._modules[0].get_schema()
            await self.build(Input(schema=input_schema))
        elif hasattr(self._modules[0], "input_schema") and len(self._modules) > 1:
            # We can build the Sequential program if the first module has the
            # `input_schema` property. This is most commonly found in Functional
            # program.
            input_schema = self._modules[0].input_schema
            await self.build(Input(schema=input_schema))

    def _lock_state(self):
        # Unlike other modules, Sequential is mutable after build.
        pass

    def _obj_type(self):
        return "Sequential"

    async def build(self, inputs):
        try:
            input_schema = standardize_schema(inputs.get_schema())
        except Exception:
            # Do not attempt to build if the program does not have a single
            # input.
            return
        if not self._modules:
            raise ValueError(
                f"Sequential program {self.name} cannot be built because it has "
                "no modules. Call `program.add(module)`."
            )
        if isinstance(self._modules[0], InputModule):
            if self._modules[0].get_schema() != input_schema:
                raise ValueError(
                    f"Sequential program '{self.name}' has already been "
                    "configured to use input schema "
                    f"{self._modules[0].get_schema()}. You cannot build it "
                    f"with input_schema {input_schema}"
                )
        else:
            self._modules = [InputModule(schema=input_schema)] + self._modules

        # Build functional program
        inputs = self._modules[0].output
        x = inputs
        for module in self._modules[1:]:
            try:
                x = await module(x)
            except NotImplementedError:
                # Can happen if spec inference is not implemented.
                # TODO: consider reverting inbound nodes on modules processed.
                return
            except TypeError as e:
                signature = inspect.signature(module.call)
                positional_args = [
                    param
                    for param in signature.parameters.values()
                    if param.default == inspect.Parameter.empty
                ]
                if len(positional_args) != 1:
                    raise ValueError(
                        "Modules added to a Sequential program "
                        "can only have a single positional argument, "
                        f"the input data model. Module {module.__class__.__name__} "
                        f"has multiple positional arguments: {positional_args}"
                    )
                raise e
        outputs = x
        self._functional = Functional(inputs=inputs, outputs=outputs)
        self.built = True

    async def call(self, inputs, training=None):
        if self._functional:
            return await self._functional.call(inputs, training=training)
        # Fallback: Just apply the module sequence.
        # This typically happens if `inputs` is a nested struct.
        for module in self.modules:
            # During each iteration, `inputs` are the inputs to `module`, and
            # `outputs` are the outputs of `module` applied to `inputs`. At the
            # end of each iteration `inputs` is set to `outputs` to prepare for
            # the next module.
            kwargs = {}
            if module._call_has_training_arg and training is not None:
                kwargs["training"] = training
            outputs = await module(inputs, **kwargs)
            inputs = outputs
        return outputs

    @property
    def modules(self):
        """Unlike Keras, also output the potentially auto-generated `InputModule`"""
        return self._modules

    @modules.setter
    def modules(self, _):
        raise AttributeError(
            "`Sequential.modules` attribute is reserved and should not be used. "
            "Use `add()` and `pop()` to change the modules in this program."
        )

    async def compute_output_spec(self, inputs, training=None):
        if self._functional:
            return await self._functional.compute_output_spec(
                inputs,
                training=training,
            )
        # Direct application
        for module in self.modules:
            outputs = await module.compute_output_spec(inputs, training=training)
            inputs = outputs
        return outputs

    @property
    def input_schema(self):
        if self._functional:
            return self._functional.input_schema
        raise AttributeError(
            f"Sequential program '{self.name}' has no defined input schema yet."
        )

    @property
    def output_schema(self):
        if self._functional:
            return self._functional.output_schema
        raise AttributeError(
            f"Sequential program '{self.name}' has no defined output schema yet."
        )

    @property
    def inputs(self):
        if self._functional:
            return self._functional.inputs
        raise AttributeError(
            f"Sequential program '{self.name}' has no defined inputs yet."
        )

    @property
    def outputs(self):
        if self._functional:
            return self._functional.outputs
        raise AttributeError(
            f"Sequential program '{self.name}' has no defined outputs yet."
        )

    def _is_module_name_unique(self, module):
        for ref_module in self._modules:
            if module.name == ref_module.name and ref_module is not module:
                return False
        return True

    def get_config(self):
        serialize_fn = serialization_lib.serialize_synalinks_object
        module_configs = []
        for module in self.modules:
            module_configs.append(serialize_fn(module))
        config = Program.get_config(self)
        config["name"] = self.name
        config["description"] = self.description
        config["modules"] = copy.deepcopy(module_configs)
        if self._functional is not None:
            config["build_input_schema"] = self._modules[0].input_schema
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        if "name" in config:
            name = config["name"]
            build_input_schema = config.get("build_input_schema")
            module_configs = config["modules"]
        else:
            name = None
            module_configs = config
        if "description" in config:
            description = config["description"]
        else:
            description = None
        program = cls(name=name, description=description)
        for module_config in module_configs:
            module = serialization_lib.deserialize_synalinks_object(
                module_config,
                custom_objects=custom_objects,
            )
            program.add(module)
        if (
            not program._functional
            and "build_input_schema" in locals()
            and build_input_schema
            and isinstance(build_input_schema, (tuple, list))
        ):
            program.build(build_input_schema)
        return program



# from synalinks.src import testing
# from synalinks.src.datasets.gsm8k import load_data


# class GSM8kTest(testing.TestCase):
#     def test_load_data(self):
#         (x_train, y_train), (x_test, y_test) = load_data()
#         self.assertTrue(len(x_train) > 0)
#         self.assertTrue(len(y_train) > 0)
#         self.assertTrue(len(x_test) > 0)
#         self.assertTrue(len(y_test) > 0)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.optimizers.greedy_optimizer import GreedyOptimizer


@synalinks_export("synalinks.optimizers.RandomFewShot")
class RandomFewShot(GreedyOptimizer):
    """Sample among the best predictions to populate the LM's prompt.

    Makes the model learn using Few Shot Learning by selecting predictions
    as examples based on their rewards using the configured sampling strategy.

    Example:

    ```python
    import synalinks
    import asyncio

    async def main():
        # ... your program definition

        program.compile(
            reward=synalinks.rewards.ExactMatch(),
            optimizer=synalinks.optimizers.RandomFewShot(
                nb_min_examples=1,
                nb_max_examples=3,
                sampling="softmax",
                sampling_temperature=1.0,
            ),
        )

        history = await program.fit(...)
    ```

    References:
        - DSPy: Compiling Declarative Language Model Calls into
          Self-Improving Pipelines (https://arxiv.org/pdf/2310.03714)
        - Language Models are Few-Shot Learners
          (https://arxiv.org/abs/2005.14165)

    Args:
        nb_min_examples (int): The min number of examples for few-shot
            learning (Default to 1).
        nb_max_examples (int): The max number of examples for few-shot
            learning (Default to 3).
        sampling (str): The method to sample predictions between
            ['random', 'best', 'softmax']. (Default to 'softmax').
        sampling_temperature (float): The temperature for softmax sampling.
            Used only when `sampling='softmax'`. Lower values concentrate
            sampling on high-reward predictions, higher values make sampling
            more uniform (Default 0.3).
        population_size (int): The maximum number of best candidates to keep
            during the optimization process.
        name (str): Optional name for the optimizer instance.
        description (str): Optional description of the optimizer instance.
    """

    def __init__(
        self,
        nb_min_examples=1,
        nb_max_examples=3,
        sampling="softmax",
        sampling_temperature=0.3,
        population_size=10,
        name=None,
        description=None,
    ):
        super().__init__(
            nb_min_examples=nb_min_examples,
            nb_max_examples=nb_max_examples,
            sampling=sampling,
            sampling_temperature=sampling_temperature,
            population_size=population_size,
            name=name,
            description=description,
        )

    async def build(self, _):
        self.built = True

    async def propose_new_candidates(
        self,
        step,
        trainable_variables,
        x=None,
        y=None,
        y_pred=None,
        training=False,
    ):
        variable_name_to_update = await self.select_variable_name_to_update(
            trainable_variables,
        )

        for trainable_variable in trainable_variables:
            if trainable_variable.name == variable_name_to_update:
                examples = await self.sample_best_predictions(
                    trainable_variable,
                )
                await self.assign_candidate(
                    trainable_variable,
                    examples=examples,
                )



# Modified from: keras/src/models/model.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import inspect
import typing
import warnings

import orjson

from synalinks.src import utils
from synalinks.src.api_export import synalinks_export
from synalinks.src.modules import Module
from synalinks.src.trainers.trainer import Trainer
from synalinks.src.utils import file_utils
from synalinks.src.utils import io_utils
from synalinks.src.utils import summary_utils


@synalinks_export(["synalinks.Program", "synalinks.programs.Program"])
class Program(Trainer, Module):
    """A program grouping modules into an object with training/inference features.

    There is four ways to instantiate a `Program`:

    ## With the "Functional API"

    You start from `Input`, you chain modules calls to specify the program's structure,
    and finally, you create your program from inputs and outputs:

    ```python
    import synalinks
    import asyncio
    
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    class AnswerWithThinking(synalinks.DataModel):
        thinking: str = synalinks.Field(
            description="Your step by step thinking process",
        )
        answer: float = synalinks.Field(
            description="The correct numerical answer",
        )
    
    async def main():

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        x0 = synalinks.Input(data_model=Query)
        x1 = await synalinks.Generator(
            data_model=AnswerWithThinking,
            language_model=language_model,
        )(x0)

        program = synalinks.Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    Note: Only dicts, lists, and tuples of input data models are supported. Nested
    inputs are not supported (e.g. lists of list or dicts of dict).

    ## By subclassing the `Program` class

    In that case, you should define your
    modules in `__init__()` and you should implement the program's structure
    in `call()` .

    ```python
    import synalinks
    import asyncio
    
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    class AnswerWithThinking(synalinks.DataModel):
        thinking: str = synalinks.Field(
            description="Your step by step thinking process",
        )
        answer: float = synalinks.Field(
            description="The correct numerical answer",
        )
        
    class ChainOfThought(synalinks.Program):
        \"\"\"Useful to answer in a step by step manner.
        
        The first line of the docstring is provided as description
        for the program if not provided in the `super().__init__()`.
        In a similar way the name is automatically infered based on
        the class name if not provided.
        \"\"\"

        def __init__(
            self,
            language_model=None,
            name=None,
            description=None,
            trainable=True,
        ):
            super().__init__(
                name=name,
                description=description,
                trainable=trainable,
            )
            self.answer = synalinks.Generator(
                data_model=AnswerWithThinking,
                language_model=language_model,
                name="generator_"+self.name,
            )

        async def call(self, inputs, training=False):
            if not inputs:
                return None
            x = await self.answer(inputs, training=training)
            return x

        def get_config(self):
            config = {
                "name": self.name,
                "description": self.description,
                "trainable": self.trainable,
            }
            language_model_config = \
            {
                "language_model": synalinks.saving.serialize_synalinks_object(
                    self.language_model
                )
            }
            return {**config, **language_model_config}

        @classmethod
        def from_config(cls, config):
            language_model = synalinks.saving.deserialize_synalinks_object(
                config.pop("language_model")
            )
            return cls(language_model=language_model, **config)
    
    async def main():

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        program = ChainOfThought(
            language_model=language_model,
        )
    ```

    If you subclass `Program`, you can optionally have
    a `training` argument (boolean) in `call()`, which you can use to specify
    a different behavior in training and inference.

    Once the program is created, you can config the program with rewards and metrics
    with `program.compile()`, train the program with `program.fit()`, or use the program
    to do prediction with `program.predict()` or `program()`.

    To understand the difference between `program.predict()` or `program()`, read the
    [FAQ](https://synalinks.github.io/synalinks/FAQ/#whats-the-difference-between-program-methods-predict-and-__call__).

    ## Mixing the subclassing and the `Functional` API

    This way of programming is recommended to encapsulate your application while 
    providing an easy to use setup. It is the recommended way for most users as 
    it avoid making your program/agents from scratch.
    In that case, you should implement only the `__init__()` and `build()` methods.

    ```python
    import synalinks
    import asyncio

    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    class AnswerWithThinking(synalinks.DataModel):
        thinking: str = synalinks.Field(
            description="Your step by step thinking process",
        )
        answer: float = synalinks.Field(
            description="The correct numerical answer",
        )

    async def main():

        class ChainOfThought(synalinks.Program):
            \"\"\"Useful to answer in a step by step manner.\"\"\"

            def __init__(
                self,
                language_model=None,
                name=None,
                description=None,
                trainable=True,
            ):
                super().__init__(
                    name=name,
                    description=description,
                    trainable=trainable,
                )

                self.language_model = language_model
            
            async def build(self, inputs):
                outputs = await synalinks.Generator(
                    data_model=AnswerWithThinking,
                    language_model=self.language_model,
                )(inputs)

                # Create your program using the functional API
                super().__init__(
                    inputs=inputs,
                    outputs=outputs,
                    name=self.name,
                    description=self.description,
                    trainable=self.trainable,
                )

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        program = ChainOfThought(
            language_model=language_model,
        )

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    This allows you to not have to implement the `call()` and serialization methods
    (`get_config()` and `from_config()`). The program will be built for any inputs 
    the first time called.

    ## With the `Sequential` class

    In addition, `synalinks.Sequential` is a special case of program where
    the program is purely a stack of single-input, single-output modules.

    ```python
    import synalinks
    import asyncio
    
    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    class AnswerWithThinking(synalinks.DataModel):
        thinking: str = synalinks.Field(
            description="Your step by step thinking process",
        )
        answer: float = synalinks.Field(
            description="The correct numerical answer",
        )

    async def main():

        language_model = synalinks.LanguageModel(model="ollama/mistral")

        program = synalinks.Sequential(
            [
                synalinks.Input(
                    data_model=Query,
                ),
                synalinks.Generator(
                    data_model=AnswerWithThinking,
                    language_model=language_model,
                ),
            ],
            name="chain_of_thought",
            description="Useful to answer in a step by step manner.",
        )

    if __name__ == "__main__":
        asyncio.run(main())
    ```
    """

    def __new__(cls, *args, **kwargs):
        # Signature detection for usage of `Program` as a `Functional`
        if functional_init_arguments(args, kwargs) and cls == Program:
            from synalinks.src.programs.functional import Functional

            return Functional.__new__(Functional, *args, **kwargs)
        return typing.cast(cls, super().__new__(cls))

    def __init__(self, *args, **kwargs):
        Trainer.__init__(self)
        from synalinks.src.programs import functional

        # Signature detection for usage of a `Program` subclass
        # as a `Functional` subclass
        if functional_init_arguments(args, kwargs):
            inject_functional_program_class(self.__class__)
            functional.Functional.__init__(self, *args, **kwargs)
        else:
            Module.__init__(self, *args, **kwargs)

    async def call(self, *args, **kwargs):
        raise NotImplementedError(
            f"Program {self.__class__.__name__} does not have a `call()` "
            "method implemented."
        )

    @property
    def modules(self):
        return list(self._flatten_modules(include_self=False, recursive=False))

    @modules.setter
    def modules(self, _):
        raise AttributeError(
            "`Program.modules` attribute is reserved and should not be used. "
            "Please use another name."
        )

    def get_module(self, name=None, index=None):
        """Retrieves a module based on either its name (unique) or index.

        If `name` and `index` are both provided, `index` will take precedence.
        Indices are based on order of horizontal graph traversal (bottom-up).

        Args:
            name (str): String, name of module.
            index (int): Integer, index of module.

        Returns:
            (Module): A module instance.
        """
        if index is not None and name is not None:
            raise ValueError(
                "Provide only a module name or a module index. Received: "
                f"index={index}, name={name}."
            )
        if index is not None:
            if len(self.modules) <= index:
                raise ValueError(
                    f"Was asked to retrieve module at index {index}"
                    f" but program only has {len(self.modules)}"
                    " modules."
                )
            else:
                return self.modules[index]

        if name is not None:
            for module in self.modules:
                if module.name == name:
                    return module
            raise ValueError(
                f"No such module: {name}. Existing modules are: "
                f"{list(module.name for module in self.modules)}."
            )
        raise ValueError("Provide either a module name or module index at `get_module`.")

    def summary(
        self,
        line_length=None,
        positions=None,
        print_fn=None,
        expand_nested=False,
        show_trainable=False,
        module_range=None,
    ):
        """Prints a string summary of the program.

        Args:
            line_length (int): Total length of printed lines
                (e.g. set this to adapt the display to different
                terminal window sizes).
            positions (list): Relative or absolute positions of log elements
                in each line. If not provided, becomes
                `[0.3, 0.6, 0.70, 1.]`. Defaults to `None`.
            print_fn (Callable): Print function to use. By default, prints to `stdout`.
                If `stdout` doesn't work in your environment, change to `print`.
                It will be called on each line of the summary.
                You can set it to a custom function
                in order to capture the string summary.
            expand_nested (bool): Whether to expand the nested models.
                Defaults to `False`.
            show_trainable (bool): Whether to show if a module is trainable.
                Defaults to `False`.
            module_range (list | tuple): a list or tuple of 2 strings,
                which is the starting module name and ending module name
                (both inclusive) indicating the range of modules to be printed
                in summary. It also accepts regex patterns instead of exact
                names. In this case, the start predicate will be
                the first element that matches `module_range[0]`
                and the end predicate will be the last element
                that matches `module_range[1]`.
                By default `None` considers all modules of the model.

        Raises:
            ValueError: if `summary()` is called before the model is built.
        """
        summary_utils.print_summary(
            self,
            line_length=line_length,
            positions=positions,
            print_fn=print_fn,
            expand_nested=expand_nested,
            show_trainable=show_trainable,
            module_range=module_range,
        )

    def save(self, filepath, overwrite=True, **kwargs):
        """Saves a program as a `.json` file.

        Example:

        ```python
        import synalinks

        class Query(synalinks.DataModel):
            query: str

        class AnswerWithRationale(synalinks.DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel("ollama/mistral")

        program = synalinks.Sequential(
            [
                synalinks.Input(data_model=Query),
                synalinks.Generator(
                    data_model=AnswerWithRationale,
                    language_model=language_model,
                ),
            ],
        )

        program.save("program.json")
        loaded_program = synalinks.programs.program_from_json("program.json")
        ```

        The saved `.json` file contains:

        - The program's configuration (architecture)
        - The program's variables
        - The program's optimizer's state (if any)
        - The program's reward's state (if any)

        Thus programs can be reinstantiated in the exact same state.

        Args:
            filepath (str | os.PathLike): `str` or `os.PathLike` object.
                The path where to save the model. Must end in `.json`.
            overwrite (bool): Whether we should overwrite any existing program at
                the target location, or instead ask the user via
                an interactive prompt. Default to `True`.
        """
        from synalinks.src.saving import serialization_lib

        filepath = file_utils.path_to_string(filepath)
        if not filepath.endswith(".json"):
            raise ValueError(
                f"The filepath should ends with '.json', received filepath={filepath}"
            )
        program_config = serialization_lib.serialize_synalinks_object(self)
        variables_config = self.get_state_tree()
        program_config.update({"variables": variables_config})
        program_config_string = orjson.dumps(
            program_config, option=orjson.OPT_INDENT_2
        ).decode()
        if file_utils.exists(filepath) and not overwrite:
            io_utils.ask_to_proceed_with_overwrite(filepath)
        with open(filepath, "w") as f:
            f.write(program_config_string)

    async def build_from_config(self, config):
        if not config:
            return
        status = False
        if "input_schema" in config:
            # Case: all inputs are in the first arg (possibly nested).
            if utils.is_default(self.build):
                status = self._build_by_run_for_single_pos_arg(config["input_schema"])
            else:
                try:
                    await self.build(config["input_schema"])
                    status = True
                except:
                    pass
            self._build_schemas_dict = config

        elif "schemas_dict" in config:
            # Case: inputs were recorded as multiple keyword arguments.
            if utils.is_default(self.build):
                status = self._build_by_run_for_kwargs(config["schemas_dict"])
            else:
                try:
                    await self.build(**config["schemas_dict"])
                    status = True
                except:
                    pass
            self._build_schemas_dict = config["schemas_dict"]

        if not status:
            warnings.warn(
                f"Program '{self.name}' had a build config, but the program "
                "cannot be built automatically in "
                "`build_from_config(config)`. "
                "You should implement "
                "`def build_from_config(self, config)`, "
                "and you might also want to implement the method "
                " that generates the config at saving time, "
                "`def get_build_config(self)`. "
                "The method `build_from_config()` is meant to "
                "create the state of the model (i.e. its variables) "
                "upon deserialization.",
                stacklevel=2,
            )

    def to_json(self, **kwargs):
        """Returns a JSON string containing the network configuration.

        ```python
        json_string = program.to_json()
        ```

        To load a network from a JSON save file, use
        `synalinks.programs.program_from_json(json_string, custom_objects={...})`.

        Args:
            **kwargs (keyword arguments): Additional keyword arguments to be passed to
                `orjson.dumps()`.

        Returns:
            (str): A JSON string.
        """
        from synalinks.src.saving import serialization_lib

        program_config = serialization_lib.serialize_synalinks_object(self)
        return orjson.dumps(program_config, **kwargs).decode()

    @classmethod
    def from_config(cls, config, custom_objects=None):
        from synalinks.src.programs.functional import Functional

        functional_config_keys = [
            "name",
            "modules",
            "input_modules",
            "output_modules",
        ]
        is_functional_config = all(key in config for key in functional_config_keys)
        argspec = inspect.getfullargspec(cls.__init__)
        functional_init_args = inspect.getfullargspec(Functional.__init__).args[1:]
        revivable_as_functional = (
            cls in {Functional, Program}
            or argspec.args[1:] == functional_init_args
            or (argspec.varargs == "args" and argspec.varkw == "kwargs")
        )
        if is_functional_config and revivable_as_functional:
            # Revive Functional model
            # (but not Functional subclasses with a custom __init__)
            from synalinks.src.programs.functional import functional_from_config

            return functional_from_config(cls, config, custom_objects=custom_objects)

        # Either the model has a custom __init__, or the config
        # does not contain all the information necessary to
        # revive a Functional model. This happens when the user creates
        # subclassed models where `get_config()` is returning
        # insufficient information to be considered a Functional model.
        # In this case, we fall back to provide all config into the
        # constructor of the class.
        try:
            return cls(**config)
        except TypeError as e:
            raise TypeError(
                "Unable to revive program from config. When overriding "
                "the `get_config()` method, make sure that the "
                "returned config contains all items used as arguments "
                f"in the  constructor to {cls}, "
                "which is the default behavior. "
                "You can override this default behavior by defining a "
                "`from_config(cls, config)` class method to specify "
                "how to create an "
                f"instance of {cls.__name__} from its config.\n\n"
                f"Received config={config}\n\n"
                f"Error encountered during deserialization: {e}"
            )

    def get_state_tree(self):
        """Retrieves tree-like structure of program variables.

        This method allows retrieval of different program variables (trainable,
        non-trainable, optimizer, and metrics). The variables are returned in a
        nested dictionary format, where the keys correspond to the variable
        names and the values are the nested representations of the variables.

        Example:

        ```python
        program.compile(
            optimizer=synalinks.optimizers.RandomFewShot(),
            reward=synalinks.rewards.ExactMatch(),
        )
        program.fit(x=x_train, y=y_train)
        state_tree = program.get_state_tree()
        ```

        Returns:
            (dict): A dictionary containing the nested representations of the
                requested variables. The keys are the variable names, and the
                values are the corresponding nested dictionaries.
        """
        variables = {}
        variables["trainable_variables"] = self._create_nested_dict(
            self.trainable_variables
        )
        variables["non_trainable_variables"] = self._create_nested_dict(
            self.non_trainable_variables
        )
        if self.optimizer:
            variables["optimizer_trainable_variables"] = self._create_nested_dict(
                self.optimizer.trainable_variables
            )
            variables["optimizer_non_trainable_variables"] = self._create_nested_dict(
                self.optimizer.non_trainable_variables
            )
        variables["metrics_variables"] = self._create_nested_dict(self.metrics_variables)
        return variables

    def _create_nested_dict(self, variables):
        flat_dict = {}
        for v in variables:
            if v.path in flat_dict:
                raise ValueError(
                    "The following variable path is found twice in the program: "
                    f"'{v.path}'. `get_state_tree()` can only be called when "
                    "all variable paths are unique. Make sure to give unique "
                    "names to your modules (and other objects)."
                )
            flat_dict[v.path] = v.get_json()

        nested_dict = {}
        for path, value in flat_dict.items():
            parts = path.split("/")
            current_dict = nested_dict
            for part in parts[:-1]:
                if part not in current_dict:
                    current_dict[part] = {}
                current_dict = current_dict[part]
            current_dict[parts[-1]] = value

        return nested_dict

    def set_state_tree(self, state_tree):
        """Assigns values to variables of the program.

        This method takes a dictionary of nested variable values, which
        represents the state tree of the program, and assigns them to the
        corresponding variables of the program. The dictionary keys represent the
        variable names (e.g., `'trainable_variables'`, `'optimizer_variables'`),
        and the values are nested dictionaries containing the variable
        paths and their corresponding values.

        Args:
            state_tree (dict): A dictionary representing the state tree of the program.
                The keys are the variable names, and the values are nested
                dictionaries representing the variable paths and their values.
        """
        for k, v in state_tree.items():
            path_value_dict = self._flatten_nested_dict(v)
            if k == "trainable_variables":
                self._assign_variable_values(self.trainable_variables, path_value_dict)
            elif k == "non_trainable_variables":
                self._assign_variable_values(
                    self.non_trainable_variables, path_value_dict
                )
            elif k == "optimizer_trainable_variables":
                if self.optimizer:
                    self._assign_variable_values(
                        self.optimizer.trainable_variables, path_value_dict
                    )
            elif k == "optimizer_non_trainable_variables":
                if self.optimizer:
                    self._assign_variable_values(
                        self.optimizer.non_trainable_variables, path_value_dict
                    )
            elif k == "metrics_variables":
                self._assign_variable_values(self.metrics_variables, path_value_dict)
            else:
                raise ValueError(f"Unknown variable name: {k}")

    def _assign_variable_values(self, variables, path_value_dict):
        for full_path, value in path_value_dict.items():
            path = "/".join(full_path.split("/")[:-1])
            field_name = full_path.split("/")[-1]
            for variable in variables:
                if variable.path == path:
                    variable.get_json()[field_name] = value

    def _flatten_nested_dict(self, nested_dict):
        flat_dict = {}

        def _flatten(current_dict, prefix=""):
            for key, value in current_dict.items():
                if isinstance(value, dict):
                    _flatten(value, prefix + key + "/")
                else:
                    flat_dict[prefix + key] = value

        _flatten(nested_dict)
        return flat_dict

    def save_variables(self, filepath, overwrite=True):
        """Saves all module variables to a `.variables.json` file.

        Args:
            filepath (str | pathlib.Path): `str` or `pathlib.Path` object.
                Path where to save the program. Must end in `.variables.json`.
            overwrite (bool): Whether we should overwrite any existing program
                at the target location, or instead ask the user
                via an interactive prompt.
        """
        filepath = file_utils.path_to_string(filepath)
        if not filepath.endswith(".variables.json"):
            raise ValueError(
                "The filepath should ends with '.variables.json', "
                f"received filepath={filepath}"
            )
        config = self.get_state_tree()
        config_string = orjson.dumps(config, option=orjson.OPT_INDENT_2).decode()
        if file_utils.exists(filepath) and not overwrite:
            io_utils.ask_to_proceed_with_overwrite(filepath)
        with open(filepath, "w") as f:
            f.write(config_string)

    def load_variables(self, filepath):
        """Load all module variables from a `.variable.json` file.

        Args:
            filepath (str | pathlib.Path): `str` or `pathlib.Path` object.
                Path to load the program's variables from.
                Must end in `.variables.json`.
        """
        filepath = file_utils.path_to_string(filepath)
        if not filepath.endswith(".variables.json"):
            raise ValueError(
                "The filepath should ends with '.variables.json', "
                f"received filepath={filepath}"
            )
        with open(filepath, "rb") as f:
            state_tree_config = orjson.loads(f.read())
        self.set_state_tree(state_tree_config)

    @classmethod
    def load(cls, filepath, custom_objects=None):
        """Load a program from a JSON file.

        Example:

        ```python
        import synalinks

        loaded_program = synalinks.Program.load("program.json")
        ```

        Args:
            filepath (str | pathlib.Path): `str` or `pathlib.Path` object.
                Path to load the program's variables from.
                Must end in `.variables.json`.
            custom_objects (dict): Optional dictionary mapping names
                (strings) to custom classes or functions to be
                considered during deserialization.

        Returns:
            (Program): A Synalinks program instance (uncompiled).
        """
        filepath = file_utils.path_to_string(filepath)
        if not filepath.endswith(".json"):
            raise ValueError(
                f"The filepath should ends with '.json', received filepath={filepath}"
            )
        with open(filepath, "r") as f:
            json_config = f.read()
        return program_from_json(json_config, custom_objects=custom_objects)


@synalinks_export("synalinks.programs.program_from_json")
def program_from_json(json_string, custom_objects=None):
    """Parses a JSON program configuration string and returns a program instance.

    Example:

    ```python
    import synalinks

    class Query(synalinks.DataModel):
        query: str

    class AnswerWithRationale(synalinks.DataModel):
        rationale: str
        answer: str

    language_model = LanguageModel("ollama/mistral")

    program = synalinks.Sequential(
        [
            synalinks.Input(data_model=Query),
            synalinks.Generator(
                data_model=AnswerWithRationale,
                language_model=language_model,
            ),
        ],
    )

    config = program.to_json()
    loaded_program = synalinks.programs.program_from_json(config)
    ```

    Args:
        json_string (str): JSON string encoding a program configuration.
        custom_objects (dict): Optional dictionary mapping names
            (strings) to custom classes or functions to be
            considered during deserialization.

    Returns:
        (Program): A Synalinks program instance (uncompiled).
    """
    from synalinks.src.saving import serialization_lib

    program_config = orjson.loads(json_string)
    variables_config = program_config.get("variables")
    program = serialization_lib.deserialize_synalinks_object(
        program_config, custom_objects=custom_objects
    )
    program.set_state_tree(variables_config)
    return program


def functional_init_arguments(args, kwargs):
    return (
        (len(args) == 2)
        or (len(args) == 1 and "outputs" in kwargs)
        or ("inputs" in kwargs and "outputs" in kwargs)
    )


def inject_functional_program_class(cls):
    """Inject `Functional` into the hierarchy of this class if needed."""
    from synalinks.src.programs import functional

    if cls is Program:
        return functional.Functional
    # In case there is any multiple inheritance, we stop injecting the
    # class if synalinks model is not in its class hierarchy.
    if cls is object:
        return object

    cls.__bases__ = tuple(inject_functional_program_class(base) for base in cls.__bases__)
    # Trigger any `__new__` class swapping that needed to happen on `Functional`
    # but did not because functional was not in the class hierarchy.
    cls.__new__(cls)

    return cls



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib
from synalinks.src.utils.async_utils import run_maybe_nested


@synalinks_export(
    [
        "synalinks.modules.UpdateKnowledge",
        "synalinks.UpdateKnowledge",
    ]
)
class UpdateKnowledge(Module):
    """Update (insert or upsert) data models in the given knowledge base.

    This module stores data models in the knowledge base, using the first
    field of the data model as the primary key for upsert operations.

    Args:
        knowledge_base (KnowledgeBase): The knowledge base to update.
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        knowledge_base=None,
        name=None,
        description=None,
        trainable=False,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        self.knowledge_base = knowledge_base

    async def _update(self, data_model):
        await self.knowledge_base.update(data_model)
        return data_model.clone(name="updated_" + data_model.name)

    async def call(self, inputs):
        if not inputs:
            return None
        outputs = tree.map_structure(
            lambda x: run_maybe_nested(self._update(x)),
            inputs,
        )
        return outputs

    async def compute_output_spec(self, inputs):
        return tree.map_structure(
            lambda x: x.clone(name="updated_" + x.name),
            inputs,
        )

    def get_config(self):
        config = {
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        knowledge_base_config = {
            "knowledge_base": serialization_lib.serialize_synalinks_object(
                self.knowledge_base
            )
        }
        return {**knowledge_base_config, **config}

    @classmethod
    def from_config(cls, config):
        knowledge_base = serialization_lib.deserialize_synalinks_object(
            config.pop("knowledge_base")
        )
        return cls(knowledge_base=knowledge_base, **config)



# Modified from: keras/src/trainers/data_adapters/array_slicing.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import math

import numpy as np

from synalinks.src import tree

ARRAY_TYPES = (np.ndarray, list, tuple)


def can_slice_array(x):
    return x is None or isinstance(x, ARRAY_TYPES) or hasattr(x, "__array__")


def train_validation_split(arrays, validation_split):
    """Split arrays into train and validation subsets in deterministic order.

    The last part of data will become validation data.

    Args:
        arrays: Data models to split. Allowed inputs are arbitrarily nested
            structures of data models of NumPy arrays.
        validation_split: Float between 0 and 1. The proportion of the dataset
            to include in the validation split. The rest of the dataset will be
            included in the training split.

    Returns:
        `(train_arrays, validation_arrays)`
    """

    flat_arrays = tree.flatten(arrays)
    unsplitable = [type(t) for t in flat_arrays if not can_slice_array(t)]
    if unsplitable:
        raise ValueError(
            "Argument `validation_split` is only supported "
            "for NumPy arrays."
            f"Found incompatible type in the input: {unsplitable}"
        )

    if all(t is None for t in flat_arrays):
        return arrays, arrays

    first_non_none = None
    for t in flat_arrays:
        if t is not None:
            first_non_none = t
            break

    # Assumes all arrays have the same batch shape or are `None`.
    batch_dim = int(first_non_none.shape[0])
    split_at = int(math.floor(batch_dim * (1.0 - validation_split)))

    if split_at == 0 or split_at == batch_dim:
        raise ValueError(
            f"Training data contains {batch_dim} samples, which is not "
            "sufficient to split it into a validation and training set as "
            f"specified by `validation_split={validation_split}`. Either "
            "provide more data, or a different value for the "
            "`validation_split` argument."
        )

    def _split(t, start, end):
        if t is None:
            return t
        return t[start:end]

    train_arrays = tree.map_structure(lambda x: _split(x, start=0, end=split_at), arrays)
    val_arrays = tree.map_structure(
        lambda x: _split(x, start=split_at, end=batch_dim), arrays
    )
    return train_arrays, val_arrays



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.backend import DataModel as DataModel
from synalinks.src.backend.common.global_state import clear_session as clear_session
from synalinks.src.backend.common.json_data_model import (
    is_json_data_model as is_json_data_model,
)
from synalinks.src.backend.common.symbolic_data_model import (
    is_symbolic_data_model as is_symbolic_data_model,
)
from synalinks.src.backend.config import api_base as api_base
from synalinks.src.backend.config import api_key as api_key
from synalinks.src.backend.config import backend as backend
from synalinks.src.backend.config import disable_telemetry as disable_telemetry
from synalinks.src.backend.config import enable_logging as enable_logging
from synalinks.src.backend.config import enable_observability as enable_observability
from synalinks.src.backend.config import epsilon as epsilon
from synalinks.src.backend.config import floatx as floatx
from synalinks.src.backend.config import get_seed as get_seed
from synalinks.src.backend.config import (
    is_observability_enabled as is_observability_enabled,
)
from synalinks.src.backend.config import is_telemetry_enabled as is_telemetry_enabled
from synalinks.src.backend.config import mlflow_experiment_name as mlflow_experiment_name
from synalinks.src.backend.config import mlflow_tracking_uri as mlflow_tracking_uri
from synalinks.src.backend.config import set_api_base as set_api_base
from synalinks.src.backend.config import set_api_key as set_api_key
from synalinks.src.backend.config import set_backend as set_backend
from synalinks.src.backend.config import set_epsilon as set_epsilon
from synalinks.src.backend.config import set_floatx as set_floatx
from synalinks.src.backend.config import set_seed as set_seed
from synalinks.src.backend.pydantic.base import ChatMessage as ChatMessage
from synalinks.src.backend.pydantic.base import ChatMessages as ChatMessages
from synalinks.src.backend.pydantic.base import ChatRole as ChatRole
from synalinks.src.backend.pydantic.base import Embedding as Embedding
from synalinks.src.backend.pydantic.base import Embeddings as Embeddings
from synalinks.src.backend.pydantic.base import GenericInputs as GenericInputs
from synalinks.src.backend.pydantic.base import GenericIO as GenericIO
from synalinks.src.backend.pydantic.base import GenericOutputs as GenericOutputs
from synalinks.src.backend.pydantic.base import GenericResult as GenericResult
from synalinks.src.backend.pydantic.base import Instructions as Instructions
from synalinks.src.backend.pydantic.base import Prediction as Prediction
from synalinks.src.backend.pydantic.base import Score as Score
from synalinks.src.backend.pydantic.base import SimilaritySearch as SimilaritySearch
from synalinks.src.backend.pydantic.base import Stamp as Stamp
from synalinks.src.backend.pydantic.base import ToolCall as ToolCalling
from synalinks.src.backend.pydantic.base import ToolCall as ToollCall
from synalinks.src.backend.pydantic.base import Trainable as Trainable
from synalinks.src.backend.pydantic.base import TripletSearch as TripletSearch
from synalinks.src.backend.pydantic.base import is_chat_message as is_chat_message
from synalinks.src.backend.pydantic.base import is_chat_messages as is_chat_messages
from synalinks.src.backend.pydantic.base import is_embedded as is_embedded
from synalinks.src.backend.pydantic.base import is_embedding as is_embedding
from synalinks.src.backend.pydantic.base import is_embeddings as is_embeddings
from synalinks.src.backend.pydantic.base import is_instructions as is_instructions
from synalinks.src.backend.pydantic.base import is_prediction as is_prediction
from synalinks.src.backend.pydantic.base import (
    is_similarity_search as is_similarity_search,
)
from synalinks.src.backend.pydantic.base import is_stamped as is_stamped
from synalinks.src.backend.pydantic.base import is_tool_call as is_tool_call
from synalinks.src.backend.pydantic.base import is_trainable as is_trainable
from synalinks.src.backend.pydantic.base import is_triplet_search as is_triplet_search
from synalinks.src.backend.pydantic.core import is_meta_class as is_meta_class
from synalinks.src.utils.naming import get_uid as get_uid
